#!/bin/sh
#installation script for QQbar_threshold and related libraries


error_exit(){
    echo $1
    exit 1
}

CMAKE="cmake .. -DCMAKE_BUILD_TYPE=Release"

#hack to convert possibly relative paths to absolute paths
if [ "$1" != "" ]; then
    if [ ! -d $1 ]; then
        printf "Usage: install.sh [DIRECTORY]\nInstall the QQbar_threshold and QCD libraries to an existing DIRECTORY\n\n"
        error_exit "Failed to install to '$1': is not a directory"
    fi
    BASEDIR=$(cd "$1"; pwd)
fi
cd $(dirname $0)

cat <<__EOF__
This script will install the QQbar_threshold and QCD libraries.
Please make sure the following programs and libraries are available.

- A compiler complying to the C++11 standard (e.g. g++ 4.8 or higher)
- cmake (http://cmake.org/)
- Odeint from the boost libraries (http://www.boost.org/)
- GSL, the GNU scientific library (http://www.gnu.org/software/gsl)

__EOF__

if [ "$BASEDIR" = "" ]; then
    BASEDIR="/usr/local/"
    else
    cat <<__EOF__
You have chosen to install to the base directory $BASEDIR.
Please make sure that all relevant files are found by your compiler and linker.
You may have to adjust environment variables such as LIBRARY_PATH,
LD_LIBRARY_PATH, and CPLUS_INCLUDE_PATH.

__EOF__
fi

echo "Press enter to continue and view the license."

read IGNORE
less COPYING || error_exit "Error while showing the license"

echo -n "Do you accept the license? [y/N]: "

read LINE LEFTOVER
if [ "$LINE" != "y" ] && [ "$LINE" != "Y" ]; then
    error_exit "License not accepted. Exiting..."
fi

CMAKE="$CMAKE -DCMAKE_INSTALL_PREFIX:PATH=$BASEDIR"

# $EUID is not guaranteed to be defined in sh? Use id -u instead.
if [ "$(id -u)" = "0" ] && command -v math >/dev/null 2>&1; then
    cat<<__EOF__
Installing Mathematica packages as root will most likely fail if you are
using a single-user Mathematica license. For installing as a different
user, please enter the user name. Note that this requires the sudo
command to be available on your machine.

Install Mathematica packages as user [root]:
__EOF__
    read MATHEMATICAUSER
    export MATHEMATICAUSER
fi

echo "Installing to $BASEDIR..."

test -d $BASEDIR/include/ || mkdir $BASEDIR/include/
test -d $BASEDIR/lib/ || mkdir $BASEDIR/lib/
export CPLUS_INCLUDE_PATH="$BASEDIR/include/:$CPLUS_INCLUDE_PATH"
export LD_LIBRARY_PATH="$BASEDIR/lib/:$LD_LIBRARY_PATH"
export LIBRARY_PATH="$BASEDIR/lib/:$LIBRARY_PATH"
# for OS X
export DYLD_LIBRARY_PATH="$BASEDIR/lib/:$DYLD_LIBRARY_PATH"

echo "Extracting sources..."
for LIB in libQCD libQQbar_threshold; do
    SOURCE=`ls ${LIB}-*-Source.tar.gz`
    SOURCE=`basename $SOURCE .tar.gz`
    tar xzf $SOURCE.tar.gz || error_exit "Failed to extract $LIB source"
done;

cat<<__EOF__

The header constants.hpp defines various physical constants like the
masses of W and Z bosons and the default settings for various
options. These values can no longer be changed after installation. If
you decide to change them now, the original settings will be saved in
the header default_constants.hpp.

Do you want to edit the predefined constants? [y/N]:
__EOF__
read CUSTOMCONSTANTS LEFTOVER

if [ "$CUSTOMCONSTANTS" = "y" ] || [ "$CUSTOMCONSTANTS" = "Y" ]; then
    CONSTANTSFILE=`find libQQbar_threshold-*-Source -name constants.hpp -print -quit`
    echo $CONSTANTSFILE
    mkdir -p "$BASEDIR/include/QQbar_threshold/"
    cp $CONSTANTSFILE "$BASEDIR/include/QQbar_threshold/default_constants.hpp"
    ${VISUAL:-${EDITOR:-vi}} $CONSTANTSFILE
fi

echo "Installing complex.hpp"
cp -v complex.hpp "$BASEDIR/include/" || error_exit "Failed to install complex.hpp"

for LIB in libQCD libQQbar_threshold; do
    echo "Installing $LIB"
    SOURCE=`ls ${LIB}-*-Source.tar.gz`
    SOURCE=`basename $SOURCE .tar.gz`
    cd $SOURCE
    test -d build || mkdir build
    cd build
    $CMAKE || error_exit "Failed to cmake $LIB"
    make || error_exit "Failed to make $LIB"
    make install || error_exit "Failed to install $LIB"
    if [ "$CUSTOMCONSTANTS" = "y" ] || [ "$CUSTOMCONSTANTS" = "Y" ]; then
        printf "Custom constants, skipping tests.\n"
        else
        make test || error_exit "Tests failed"
    fi
    cd ../..
    echo "Removing $SOURCE..."
    rm -rf $SOURCE
done
platform=$(uname)
if [ "$platform" = "Darwin" ]; then
    printf "Adjusting paths\n"
    ABSLIBDIR=$(cd "$BASEDIR/lib"; pwd)
    install_name_tool -change libQCD.dylib "$ABSLIBDIR/libQCD.dylib" "$ABSLIBDIR/libQQbar_threshold.dylib"
fi
echo "Done."
