/**
 * @file
 * @brief Decoupling constant for the strong coupling constant.
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2014 Andreas Maier
 *
 * This file is part of the QCD library.
 *
 * The QCD library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QCD library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QCD library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <array>

#include "scheme.hpp"

#if __cplusplus >= 201402L
#define QCD_CONSTEXPR_AFTER_CXX14 constexpr
#else
#define QCD_CONSTEXPR_AFTER_CXX14
#endif

//!
namespace QCD{

  // decoupling coefficients for alpha_s from
  // Chetyrkin, Kniehl, Steinhauser hep-ph/9708255
  // Chetyrkin, Kuehn, Steinhauser  hep-ph/0004189
  // zeta_alpha_s = zeta_g^2
  // i.e.
  // alpha_s^(nf-1) = zeta_alpha_s * alpha_s^(nf)
  // alpha_s^(nf+1) = zeta_alpha_s_inv * alpha_s^(nf)
  // lm = log(mu^2/m^2)

  constexpr unsigned zeta_alpha_s_max_order = 3;

  namespace detail{

    namespace zeta_alpha_s{
      constexpr double pi = 3.141592653589793238462643383279;

      constexpr double zeta_2 = pi*pi/6.;
      constexpr double zeta_3 = 1.202056903159594285399738161511449990765;
      constexpr double log_2 = 0.693147180559945309428690474185;

      constexpr unsigned max_nf_pow(unsigned order){
	return (order == 3u)?1u:0u;
      }

      constexpr unsigned max_log_pow(unsigned order, unsigned nf_pow){
	return (nf_pow == 0u)?order:2u;
      }

      // coefficients of zeta_alpha_s
      constexpr size_t num_coeff = 13;
      constexpr auto data_OS = std::array<double, num_coeff>{{
	  + 1., //alpha_s^0
	  + 0., - 1./6., //alpha_s^1 log^0...1
	  - 7./24., - 19./24., 1./36., //alpha_s^2 log^0...2
	  - 68849./124416. - 7./9.*zeta_2 - 2./9.*zeta_2*log_2
	  - 80507./27648.*zeta_3, //alpha_s^3 nf^0 log^0
	  - 4465./864., - 131./576., - 1./216., //alpha_s^3 nf^0 log^1...3
	  + 2479./31104. + zeta_2/9, + 409./1728., 0. //alpha_s^3 nf^1 log^0...2
	}};

      constexpr auto data_MS = std::array<double, num_coeff>{{
	  1.,
	  0., -1./6.,
	  11./72., -11./24., 1./36.,
	  575263./124416. - 82043./27648.*zeta_3,
	  -511./288., 23./192., -1./216.,
	  -2633./31104., 67./576., -1./36
	}};

      constexpr auto data_SI = std::array<double, num_coeff>{{
	  1.,
	  0., -1./6.,
	  11./72., -19./24., 1./36.,
	  575263./124416 - 82043./27648.*zeta_3,
	  -131./32., -131./576., -1./216.,
	  -2633./31104., 281./1728., 0.
	}};

      template<scheme s>
      QCD_CONSTEXPR_AFTER_CXX14
      double coeffs(size_t);

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs<scheme::pole>(size_t i){
	return data_OS[i];
      }

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs<scheme::MSbar>(size_t i){
	return data_MS[i];
      }

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs<scheme::scale_invariant>(size_t i){
	return data_SI[i];
      }

      // coefficients of zeta_alpha_s^-1
      constexpr auto data_OS_inv = std::array<double, num_coeff>{{
	  + 1., //alpha_s^0
	  + 0., 1./6., //alpha_s^1 log^0...1
	  + 7./24., + 19./24., 1./36., //alpha_s^2 log^0...2
	  + 58933./124416. + 2./3.*zeta_2 + 2./9.*zeta_2*log_2
	  + 80507./27648.*zeta_3, //alpha_s^3 nf^0 log^0
	  + 8941./1728., + 511./576., + 1./216., //alpha_s^3 nf^0 log^1...3
	  - 2479./31104. - zeta_2/9, - 409./1728., 0.0 //alpha_s^3 nf^1 log^0...2
	}};


      constexpr auto data_MS_inv = std::array<double, num_coeff>{{
	  + 1., //alpha_s^0
	  + 0., 1./6., //alpha_s^1 log^0...1
	  - 11./72., + 11./24., 1./36., //alpha_s^2 log^0...2
	  - 564731./124416. + 82043./27648.*zeta_3, //alpha_s^3 nf^0 log^0
	  + 2645./1728., + 167./576., + 1./216., //alpha_s^3 nf^0 log^1...3
	  + 2633./31104., - 67./576., 1./36. //alpha_s^3 nf^1 log^0...2
	}};

      constexpr auto data_SI_inv = std::array<double, num_coeff>{{
	  + 1., //alpha_s^0
	  + 0., 1./6., //alpha_s^1 log^0...1
	  - 11./72., + 19./24., 1./36., //alpha_s^2 log^0...2
	  - 564731./124416. + 82043./27648.*zeta_3, //alpha_s^3 nf^0 log^0
	  + 2191./576., + 511./576., + 1./216., //alpha_s^3 nf^0 log^1...3
	  + 2633./31104., - 281./1728., 0. //alpha_s^3 nf^1 log^0...2
	}};

      template<scheme s>
      QCD_CONSTEXPR_AFTER_CXX14
      double coeffs_inv(size_t);

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs_inv<scheme::pole>(size_t i){
	return data_OS_inv[i];
      }

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs_inv<scheme::MSbar>(size_t i){
	return data_MS_inv[i];
      }

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs_inv<scheme::scale_invariant>(size_t i){
	return data_SI_inv[i];
      }


      // get index in coeffs array for a given order and power of nf, lm
      constexpr unsigned get_index(
	  unsigned order,
	  unsigned nf_pow
      );

      constexpr unsigned get_index(
	  unsigned order
      ){
	return (order == 0u)?
	  0:(
	  + 1u
	  + max_log_pow(order - 1, max_nf_pow(order - 1))
	  + get_index(order - 1, max_nf_pow(order - 1))
	  );
      }

      constexpr unsigned get_index(
	  unsigned order,
	  unsigned nf_pow
      ){
	return (nf_pow == 0u)?
	  get_index(order):(
	  1u + max_log_pow(order, nf_pow - 1) + get_index(order, nf_pow - 1)
	  );
      }

      constexpr unsigned get_index(
	  unsigned order,
	  unsigned nf_pow,
	  unsigned log_pow
      ){
	return log_pow + get_index(order, nf_pow);
      }
    }
  }

  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_alpha_s_log_coeff(
      unsigned order,
      unsigned nf_pow,
      unsigned log_pow
  ){
    using namespace detail::zeta_alpha_s;
    return coeffs<s>(get_index(order, nf_pow, log_pow));
  }

  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_alpha_s_inv_log_coeff(
      unsigned order,
      unsigned nf_pow,
      unsigned log_pow
  ){
    using namespace detail::zeta_alpha_s;
    return coeffs_inv<s>(get_index(order, nf_pow, log_pow));
  }


  namespace detail{
    namespace zeta_alpha_s{

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double nf_coeff_helper(
	  unsigned order,
	  unsigned nf_pow,
	  double lm,
	  unsigned cur_log_pow
      ){
	return (cur_log_pow == max_log_pow(order, nf_pow))?
	  zeta_alpha_s_log_coeff<s>(order, nf_pow, cur_log_pow):(
	      + zeta_alpha_s_log_coeff<s>(order, nf_pow, cur_log_pow)
	      + lm * nf_coeff_helper<s>(order, nf_pow, lm, cur_log_pow + 1)
	  );
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double nf_coeff_helper_inv(
	  unsigned order,
	  unsigned nf_pow,
	  double lm,
	  unsigned cur_log_pow
      ){
	return (cur_log_pow == max_log_pow(order, nf_pow))?
	  zeta_alpha_s_inv_log_coeff<s>(order, nf_pow, cur_log_pow):(
	      + zeta_alpha_s_inv_log_coeff<s>(order, nf_pow, cur_log_pow)
	      + lm * nf_coeff_helper_inv<s>(order, nf_pow, lm, cur_log_pow + 1)
	  );
      }
    }
  }

  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_alpha_s_nf_coeff(
      unsigned order,
      unsigned nf_pow,
      double lm
  ){
    using namespace detail::zeta_alpha_s;
    return nf_coeff_helper<s>(order, nf_pow, lm, 0);
  }

  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_alpha_s_inv_nf_coeff(
      unsigned order,
      unsigned nf_pow,
      double lm
  ){
    using namespace detail::zeta_alpha_s;
    return nf_coeff_helper_inv<s>(order, nf_pow, lm, 0);
  }

  namespace detail{
    namespace zeta_alpha_s{
      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper(
	  unsigned order,
	  unsigned nf,
	  double lm,
	  unsigned cur_nf_pow
      ){
	return (cur_nf_pow == max_nf_pow(order))?
	  zeta_alpha_s_nf_coeff<s>(order, cur_nf_pow, lm):(
	      + zeta_alpha_s_nf_coeff<s>(order, cur_nf_pow, lm)
	      + nf * helper<s>(order, nf, lm, cur_nf_pow + 1)
	  );
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper_inv(
	  unsigned order,
	  unsigned nf,
	  double lm,
	  unsigned cur_nf_pow
      ){
	return (cur_nf_pow == max_nf_pow(order))?
	  zeta_alpha_s_inv_nf_coeff<s>(order, cur_nf_pow, lm):(
	      + zeta_alpha_s_inv_nf_coeff<s>(order, cur_nf_pow, lm)
	      + nf * helper_inv<s>(order, nf, lm, cur_nf_pow + 1)
	  );
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double zeta_alpha_s(
	  unsigned order,
	  unsigned nf,
	  double lm
      ){
	return helper<s>(order, nf, lm, 0);
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double zeta_alpha_s_inv(
	  unsigned order,
	  unsigned nf,
	  double lm
      ){
	return helper_inv<s>(order, nf, lm, 0);
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper(
	  unsigned order,
	  unsigned nf,
	  double lm,
	  double alpha_s,
	  unsigned cur_order
      ){
	return (cur_order == order)?
	  zeta_alpha_s<s>(order, nf, lm):(
              + zeta_alpha_s<s>(cur_order, nf, lm)
              + alpha_s/pi * helper<s>(order, nf, lm, alpha_s, cur_order + 1)
	  );
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper_inv(
	  unsigned order,
	  unsigned nf,
	  double lm,
	  double alpha_s,
	  unsigned cur_order
      ){
	return (cur_order == order)?
	  zeta_alpha_s_inv<s>(order, nf, lm):(
              + zeta_alpha_s_inv<s>(cur_order, nf, lm)
              + alpha_s/pi * helper_inv<s>(order, nf, lm, alpha_s, cur_order + 1)
	  );
      }
    }
  }

  //!Decoupling constant for the strong coupling constant.
  /**
   * @tparam s      Decoupling scheme
   * @param order   Perturbative order
   * @param nf      Number of active flavours before decoupling
   * @param lm      \f$\ln(\mu^2/m^2)\f$, where \f$\mu\f$ is the decoupling scale and \f$m\f$ the mass of the decoupled quark
   * @param alpha_s strong coupling constant at the decoupling scale
   * @returns       The decoupling constant to compute the strong coupling in the theory with nf-1  flavours
   *
   * The decoupling constant \f$\zeta_{\alpha_s}\f$ is defined by
   * \f$\alpha_s^{(n_f-1)} = \zeta_{\alpha_s} \alpha_s^{(n_f)}\f$,
   * where \f$\alpha_s^{(n_f)}\f$ is the coupling constant in the theory
   * with $n_f$ active flavours.
   *
   * All input parameters are defined in the theory with nf active flavours.
   */
  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_alpha_s(
      unsigned order,
      unsigned nf,
      double lm,
      double alpha_s
  ){
    using namespace detail::zeta_alpha_s;
    return helper<s>(order, nf, lm, alpha_s, 0);
  }

  //!Inverse decoupling constant for the strong coupling constant.
  /**
   * @tparam s      Decoupling scheme
   * @param order   Perturbative order
   * @param nf      Number of active flavours before decoupling
   * @param lm      \f$\ln(\mu^2/m^2)\f$, where \f$\mu\f$ is the decoupling scale and \f$m\f$ the mass of the decoupled quark
   * @param alpha_s strong coupling constant with nf active flavours at the decoupling scale
   * @returns       The decoupling constant to compute the strong coupling in the theory with nf+1  flavours
   *
   * The inverse decoupling constant \f$\zeta_{\alpha_s}^{-1}\f$ is defined by
   * \f$\alpha_s^{(n_f+1)} = \zeta_{\alpha_s}^{-1} \alpha_s^{(n_f)}\f$,
   * where \f$\alpha_s^{(n_f)}\f$ is the coupling constant in the theory
   * with $n_f$ active flavours.
   *
   */
  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_alpha_s_inv(
      unsigned order,
      unsigned nf,
      double lm,
      double alpha_s
  ){
    using namespace detail::zeta_alpha_s;
    return helper_inv<s>(order, nf, lm, alpha_s, 0);
  }

  //!Coefficients of the decoupling constant.
  /**
   * @tparam s      Decoupling scheme
   * @param order   Perturbative order
   * @param nf      Number of active flavours before decoupling
   * @param lm      \f$\ln(\mu^2/m^2)\f$, where \f$\mu\f$ is the decoupling scale and \f$m\f$ the mass of the decoupled quark
   * @returns       The perturbative coefficient of the decoupling constant at the given order
   *
   * The coefficients \f$\zeta_{\alpha_s, i}\f$ are defined via
   * \f$\zeta_{\alpha_s} = \sum_{i=0}^\infty \zeta_{\alpha_s, i} (\frac{\alpha_s}{\pi})^i\f$.
   */
  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_alpha_s(
      unsigned order,
      unsigned nf,
      double lm
  ){
    return detail::zeta_alpha_s::zeta_alpha_s<s>(order, nf, lm);
  }

  //!Coefficients of the inverse decoupling constant.
  /**
   * @tparam s      Decoupling scheme
   * @param order   Perturbative order
   * @param nf      Number of active flavours before decoupling
   * @param lm      \f$\ln(\mu^2/m^2)\f$, where \f$\mu\f$ is the decoupling scale and \f$m\f$ the mass of the decoupled quark
   * @returns       The perturbative coefficient of the inverse decoupling constant at the given order
   *
   * The coefficients \f$\zeta_{\alpha_s, i}^{-1}\f$ are defined via
   * \f$\zeta_{\alpha_s}^{-1} = \sum_{i=0}^\infty \zeta_{\alpha_s, i}^{-1} (\frac{\alpha_s}{\pi})^i\f$.
   */
  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_alpha_s_inv(
      unsigned order,
      unsigned nf,
      double lm
  ){
    return detail::zeta_alpha_s::zeta_alpha_s_inv<s>(order, nf, lm);
  }

}

#undef QCD_CONSTEXPR_AFTER_CXX14
