/**
 * @file
 * @brief ---
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QCD library.
 *
 * The QCD library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QCD library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QCD library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include <cassert>
#include <cmath>
#include <vector>
#include <array>
#include <stdexcept>
#include <string>

#include <boost/numeric/odeint.hpp>

#include "alpha_s.hpp"
#include "beta.hpp"
#include "zeta_alpha_s.hpp"

namespace QCD{

  namespace{
  }

  strong_coupling::strong_coupling(
      reference_coupling ref,
      int order
  ): alpha_s_mu0(ref.alpha_s), mu0(ref.mu), nf0(ref.nf), running_order(order) {
    if(order < 0 || order > max_running_order){
      throw std::out_of_range(
          "order " + std::to_string(order) + " not in range 0..."
          + std::to_string(max_running_order)
      );
    }
  }

  double strong_coupling::operator()(double mu) const{
    return evolve(alpha_s_mu0, mu0, mu, nf0);
  }

  int strong_coupling::order() const{
    return running_order;
  }

  int strong_coupling::nf() const{
    return nf0;
  }

  namespace{
    typedef std::array<double,1> state_type;

    class strong_coupling_RGE{
      std::vector<double> beta_coeffs;

      double beta(double alpha_s) const{
	static constexpr double pi = 3.141592653589793238462643383279;
	assert(beta_coeffs.size()>0);
	const double as = alpha_s/pi;
	double result = 0;
	for(unsigned order = beta_coeffs.size()-1; order > 0;--order){
	  result += beta_coeffs[order];
	  result *= as;
	}
	return( -as*(result + beta_coeffs[0]));
      }

    public:
      strong_coupling_RGE(int nf, int order)
	: beta_coeffs(std::vector<double>(order+1))
      {
	for(int i = 0; i <= order; ++i){
	  beta_coeffs[i] = QCD::beta(i,nf);
	}
      }

      void operator()(
		      state_type const & alpha_s,
		      state_type & dalpha_s,  //dalpha_s[0] = d/d(log(mu^2)) alpha_s
		      double const // log(mu^2)
		      ) const{
	dalpha_s[0] = alpha_s[0] * beta(alpha_s[0]);
      }
    };


  }
  //evolve alpha_s(mu0) to mu (without decoupling)
  double
  strong_coupling::evolve(double alpha_s, double mu0, double mu, int nf) const{
    using boost::numeric::odeint::integrate;
    auto RGE = strong_coupling_RGE{nf, order()};
    auto alpha_s_cur = state_type{{alpha_s}};
    double const start = 2*log(mu0);
    double const end = 2*log(mu);
    double const step_size = (end-start)/100.;
    integrate(RGE, alpha_s_cur, start, end, step_size);
    return alpha_s_cur[0];
  }

  double
  strong_coupling::decouple_up(double alpha_s, threshold thr, int nf) const{
    switch(thr.dec_scheme){
    case scheme::pole:
      return alpha_s*zeta_alpha_s_inv<scheme::pole>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::MSbar:
      return alpha_s*zeta_alpha_s_inv<scheme::MSbar>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::scale_invariant:
      return alpha_s*zeta_alpha_s_inv<scheme::scale_invariant>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    default:
      throw std::invalid_argument("decoupling not implemented for scheme");
    }
  }

  double
  strong_coupling::decouple_down(double alpha_s, threshold thr, int nf) const{
    switch(thr.dec_scheme){
    case scheme::pole:
      return alpha_s*zeta_alpha_s<scheme::pole>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::MSbar:
      return alpha_s*zeta_alpha_s<scheme::MSbar>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::scale_invariant:
      return alpha_s*zeta_alpha_s<scheme::scale_invariant>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    default:
      throw std::invalid_argument("decoupling not implemented for scheme");
    }
  }

  strong_coupling
  strong_coupling::remove(threshold thr) const{
    return decouple_down(thr);
  };

  strong_coupling
  strong_coupling::decouple_down(threshold thr) const{
    strong_coupling const & alpha_s_nf = *this;
    strong_coupling alpha_s_nfm1 = alpha_s_nf;
    alpha_s_nfm1.mu0 = thr.mu;
    alpha_s_nfm1.alpha_s_mu0 = decouple_down(alpha_s_nf(thr.mu), thr, nf0);
    --alpha_s_nfm1.nf0;
    return alpha_s_nfm1;
  }

  strong_coupling
  strong_coupling::add(threshold thr) const{
    return decouple_up(thr);
  };

  strong_coupling strong_coupling::decouple_up(threshold thr) const{
    strong_coupling const & alpha_s_nf = *this;
    strong_coupling alpha_s_nfp1 = alpha_s_nf;
    alpha_s_nfp1.mu0 = thr.mu;
    alpha_s_nfp1.alpha_s_mu0 = decouple_up(alpha_s_nf(thr.mu), thr, nf0);
    ++alpha_s_nfp1.nf0;
    return alpha_s_nfp1;
  }

}
