#include <cmath>
#include <limits>
#include <iostream>

#include "QQbar_threshold/load_grid.hpp"
#include "QQbar_threshold/xsection.hpp"
#include "QQbar_threshold/threshold.hpp"
#include "QQbar_threshold/residues.hpp"
#include "QQbar_threshold/energy_levels.hpp"
#include "QQbar_threshold/scheme_conversion.hpp"
#include "QQbar_threshold/integrate.hpp"

namespace QQt = QQbar_threshold;

static constexpr double pi = 3.14159265358979323846264338328;

double Z(int N, double mb_PS){
  return QQt::bbbar_residue(N, mb_PS, mb_PS, QQt::N3LO);
}

double E(int N, double mb_PS){
  return QQt::bbbar_energy_level(N, mb_PS, mb_PS, QQt::N3LO);
}

double Rb(double s, double mb_PS){
  try{
    return QQt::bbbar_R_ratio(std::sqrt(s), mb_PS, mb_PS, QQt::N3LO);
  }
  catch(std::out_of_range){
    return std::numeric_limits<double>::quiet_NaN();
  }
}

double M_resonances(int n, double mb_PS){
  static constexpr int N_c = 3;
  static constexpr double e_b = QQt::e_d;

  double sum_N = 0.0;
  for(int N = 1; N <= 6; ++ N){
    sum_N += Z(N, mb_PS)*std::pow(2*mb_PS + E(N, mb_PS), -2*n-1);
  }
  const double mb_pole = QQt::bottom_pole_mass(mb_PS, mb_PS, QQt::N3LO);
  return 12*pi*pi*N_c*e_b*e_b/(mb_pole*mb_pole)*sum_N;
}

double M_continuum(int n, double mb_PS){
  double const s0 =
    pow(QQt::bbbar_threshold(mb_PS, mb_PS, QQt::N3LO), 2);
  auto s = [=](double x){
    return s0 + x/(1-x);
  };
  auto integrand = [=](double x){
    return Rb(s(x), mb_PS)*std::pow(s(x), -n-1)*std::pow(1 - x, -2);
  };
  return QQt::integrate(integrand, 0, 1);
}

double M(int n, double mb_PS){
  return pow(10, 2*n)*(M_resonances(n, mb_PS) + M_continuum(n, mb_PS));
}

int main(){
  QQt::load_grid(QQt::grid_directory() + "bbbar_grid_large.tsv");
  std::cout << M(10, 4.5322) << '\n';
}
