/**
 * @file
 * @brief Insertion sums provided by grids.
 *
 * This header contains those sums appearing in the potential insertions that
 * have to obtained via interpolation from a grid generated with the
 * TTbarGridCalc.m Mathematica program. To this end, each insertion sum
 * has an associated matrix (declared in insertion_matrices.hpp). Interpolation
 * is then performed between the entries of this matrix.
 *
 * For the convention behind the naming of insertions see insertions.hpp
 *
 * @see insertions.hpp insertion_matrices.hpp, insertion_matrices.cpp
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <complex>
#include <algorithm>
#include <string>
#include <exception>

#include "parameters.hpp"

namespace QQbar_threshold{
  namespace detail{
    typedef std::complex<double> complex;

    enum{
      ep = 1000,
      delta = 100,
      one_half = 10
    };
    typedef int ins_type;

    template<ins_type I>
    complex insertion_sum1(complex lam);

    template<ins_type I>
    complex insertion_sum2(complex lam);

    template<ins_type I1, ins_type I2>
    complex insertion_sum1(complex lam);

    template<ins_type I1, ins_type I2>
    complex insertion_sum2(complex lam);

    template<ins_type I1, ins_type I2>
    complex insertion_sum3(complex lam);

    template<ins_type I1, ins_type I2, ins_type I3>
    complex insertion_sum1(complex lam);

    template<ins_type I1, ins_type I2, ins_type I3>
    complex insertion_sum2(complex lam);

    template<ins_type I1, ins_type I2, ins_type I3>
    complex insertion_sum3(complex lam);

    template<ins_type I>
    complex der_insertion_sum1(complex lam);

    template<ins_type I1, ins_type I2>
    complex der_insertion_sum1(complex lam);

    enum insertion_sum{
      insertion_sum1_1__2 = 0,
      insertion_sum1_1__3,
      insertion_sum1_1_1__0_1,
      insertion_sum1_1_1__1_1,
      insertion_sum1_1_1__0_2,
      insertion_sum2_1_1__0_2,
      insertion_sum1_1_1__1_2,
      insertion_sum2_1_1__1_2,
      insertion_sum3_1_1__1_2,
      insertion_sum1_1_1_1__0_0_1,
      insertion_sum1_1_1_1__0_1_0,
      insertion_sum1_1_1_1__1_0_1,
      insertion_sum1_1_1_1__0_1_1,
      insertion_sum2_1_1_1__0_1_1,
      insertion_sum1_1_1_1__1_1_1,
      insertion_sum2_1_1_1__1_1_1,
      insertion_sum3_1_1_1__1_1_1,
      insertion_sum1_1_one_half__1_0,
      insertion_sum2_1_one_half__1_0,
      insertion_sum1_one_half__1,
      insertion_sum1_1_delta__1_0,
      insertion_sum1_0__1,
      insertion_sum2_0__1,
      insertion_US_top,
      der_insertion_sum1_1__2,
      der_insertion_sum1_1_1__0_1,
      der_insertion_sum1_1_1__1_1,
      last_insertion_sum = der_insertion_sum1_1_1__1_1
    };

    complex get_interpolation(complex lam, insertion_sum ins);

    template<>
    inline
    complex insertion_sum1<1+2*ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_1__2);
    }

    template<>
    inline
    complex insertion_sum1<1+3*ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_1__3);
    }

    template<>
    inline
    complex insertion_sum1<1, 1+ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_1_1__0_1);
    }

    template<>
    inline
    complex insertion_sum1<1+ep, 1+ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_1_1__1_1);
    }

    template<>
    inline
    complex insertion_sum1<1, 1+2*ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_1_1__0_2);
    }

    template<>
    inline
    complex insertion_sum2<1, 1+2*ep>(complex lam){
      return get_interpolation(lam, insertion_sum2_1_1__0_2);
    }

    template<>
    inline
    complex insertion_sum1<1+ep, 1+2*ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_1_1__1_2);
    }

    template<>
    inline
    complex insertion_sum2<1+ep, 1+2*ep>(complex lam){
      return get_interpolation(lam, insertion_sum2_1_1__1_2);
    }

    template<>
    inline
    complex insertion_sum3<1+ep, 1+2*ep>(complex lam){
      return get_interpolation(lam, insertion_sum3_1_1__1_2);
    }

    template<>
    inline
    complex insertion_sum1<1, 1, 1+ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_1_1_1__0_0_1);
    }

    template<>
    inline
    complex insertion_sum1<1, 1+ep, 1>(complex lam){
      return get_interpolation(lam, insertion_sum1_1_1_1__0_1_0);
    }

    template<>
    inline
    complex insertion_sum1<1+ep, 1, 1+ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_1_1_1__1_0_1);
    }

    template<>
    inline
    complex insertion_sum1<1, 1+ep, 1+ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_1_1_1__0_1_1);
    }

    template<>
    inline
    complex insertion_sum2<1, 1+ep, 1+ep>(complex lam){
      return get_interpolation(lam, insertion_sum2_1_1_1__0_1_1);
    }

    template<>
    inline
    complex insertion_sum1<1+ep, 1+ep, 1+ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_1_1_1__1_1_1);
    }

    template<>
    inline
    complex insertion_sum2<1+ep, 1+ep, 1+ep>(complex lam){
      return get_interpolation(lam, insertion_sum2_1_1_1__1_1_1);
    }

    template<>
    inline
    complex insertion_sum3<1+ep, 1+ep, 1+ep>(complex lam){
      return get_interpolation(lam, insertion_sum3_1_1_1__1_1_1);
    }

    template<>
    inline
    complex insertion_sum1<1+ep, one_half>(complex lam){
      return get_interpolation(lam, insertion_sum1_1_one_half__1_0);
    }

    template<>
    inline
    complex insertion_sum2<1+ep, one_half>(complex lam){
      return get_interpolation(lam, insertion_sum2_1_one_half__1_0);
    }

    template<>
    inline
    complex insertion_sum1<one_half+ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_one_half__1);
    }

    template<>
    inline
    complex insertion_sum1<1+ep, delta>(complex lam){
      return get_interpolation(lam, insertion_sum1_1_delta__1_0);
    }

    template<>
    inline
    complex insertion_sum1<ep>(complex lam){
      return get_interpolation(lam, insertion_sum1_0__1);
    }

    template<>
    inline
    complex insertion_sum2<ep>(complex lam){
      return get_interpolation(lam, insertion_sum2_0__1);
    }

    complex insertion_US(complex lam);

    template<>
    inline
    complex der_insertion_sum1<1+2*ep>(complex lam){
      return get_interpolation(lam, der_insertion_sum1_1__2);
    }

    template<>
    inline
    complex der_insertion_sum1<1, 1+ep>(complex lam){
      return get_interpolation(lam, der_insertion_sum1_1_1__0_1);
    }

    template<>
    inline
    complex der_insertion_sum1<1+ep, 1+ep>(complex lam){
      return get_interpolation(lam, der_insertion_sum1_1_1__1_1);
    }

  }
}
