/**
 * @file
 * @brief P-wave corrections to Green function, Energy levels, and residues
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once
#include <complex>
#include <cmath>

#include "complex.hpp"
#include "QCD/colour.hpp"
#include "QCD/beta.hpp"
#include "harmonic_sums.hpp"
#include "hypergeometric.hpp"
#include "log_lam.hpp"
#include "log_n.hpp"
#include "zeta.hpp"
#include "psi.hpp"
#include "parameters.hpp"
#include "Wilson_coefficients.hpp"

namespace QQbar_threshold{
  using namespace complex;
  using QCD::c_F;
  using QCD::beta;
  using harmonic_sums::harmonic_sum;
  using zeta::pi;
  using psi::psi;
  typedef std::complex<double> complex;

  //TODO: move stuff into .cpp file
  // for definitions see Beneke, Piclum, Rauh arXiv:1312.4792

  /***************************************************/
  /*                Green functions                  */
  /***************************************************/

  namespace p_wave{
    using QQbar_threshold::detail::log_lam;

    //! Insertions to P-wave Green functions
    /**
     * @param lam     The value of \f$\lambda = \frac{\alpha_s C_F}{2} \sqrt{-\frac{m}{E + i\Gamma}}$\f.
     * @param p       Further parameters.
     *
     * For notation, see insertions.hpp
     */
    template<int qpow, int eppow>
    complex Insertion_P(complex lam, parameters const & p);

    template<>
    complex Insertion_P<1,0>(complex lam, parameters const & p){
      return
	- (1 - 3*pow(lam,2))*log_lam(lam, p.mu_w, p)
	+ 2 + 3*lam - 26./5.*pow(lam,2)
	+ (3*pow(lam,2) - 1)*harmonic_sum<1>(1-lam)
	+ (lam - pow(lam,3))*psi<1>(2 - lam)
	;
    }

    template<>
    complex Insertion_P<1,1>(complex lam, parameters const & p){
      using hypergeometric::PFQ;
      return (
	  + (-1 + 3*pow(lam, 2))*pow(log_lam(lam, p.mu_w, p),2)
	  + (2 - 6*pow(lam, 2))*log_lam(lam, p.mu_w, p)*log_lam(lam, p)
	  - 1./2.*pow(lam, 2)*log_lam(lam, p.mu_w, p)
	  - (4 + 6*lam - 21./2.*pow(lam, 2))*log_lam(lam, p)
	  + 59./9. + 5./72.*pi*pi + 9*lam
	  - (344./15. + pi*pi/8)*pow(lam,2)
	  + (
	      -4 - 6*lam + 10*pow(lam,2) - 2*(3*pow(lam,2) - 1)*log_lam(lam, p)
	  )* harmonic_sum<1>(1-lam)
	  + (3*pow(lam,2) - 1)*(3*psi<1>(2 - lam) - pow(harmonic_sum<1>(1-lam),2))
	  + 3/(4 - 2*lam)*PFQ<4>({1,1,4,4},{5,5,3 - lam},1)
	  + (lam - pow(lam,3))*(
	      + psi<1>(2 - lam)*(
		  + 22./3. - 2*log_lam(lam, p) - 2*harmonic_sum<1>(1-lam)
	      )
	      + psi<2>(2 - lam)
	  )
      );
    }
  }

  //! Insertions to P-wave Green functions
  /**
     * @tparam N      Perturbative order. Leading order is N = 0.
     * @param lam     The value of \f$\lambda = \frac{\alpha_s C_F}{2} \sqrt{-\frac{m}{E + i\Gamma}}$\f.
     * @param p       Further parameters.
     * @returns       The Nth order correction to the P-wave Green function.
   */
  template <int N>
  complex Green_function_P(complex lam, parameters const & p);

  template<>
  complex Green_function_P<0>(complex lam, parameters const & p){
    using namespace p_wave;
    return  (
        -1/pi*p.m*pow(p.m*c_F*p.alpha_s/(4*lam), 3)*(
            -2 + lam*(8 + lam*(6 - 7*lam))
            + 4*lam*(pow(lam,2)-1) * (
                harmonic_sum<1>(1-lam) + log_lam(lam, p.mu_w, p)
            )
        )
    );
  }

  template<>
  complex Green_function_P<1>(complex lam, parameters const & p){
    using namespace p_wave;
    using Wilson_coefficients::a1_eff;
    using Wilson_coefficients::a_ep;
    return
      p.contributions.v_Coulomb[0]*(
	  -pow(c_F,3)*pow(p.m,4)*pow(p.alpha_s,4)/64./pow(pi,2)/pow(lam,2)
      )*(
	  + a1_eff(p)*Insertion_P<1, 0>(lam, p)
	  + a_ep<1>(p.nl)*(1./6. - 3./10.*pow(lam,2))
	  + 4*beta(0, p.nl)*Insertion_P<1, 1>(lam, p)
      );
  }

  /***************************************************/
  /*              poles (energy levels)              */
  /***************************************************/

  //! Leading-order P-wave binding energy
  /**
   * @param n    Principal quantum number.
   * @param p    Further parameters.
   * @return     The binding energy
   */
  inline double E_n_P_LO(int n, parameters const & p){
    return -p.m*pow(c_F*p.alpha_s/(2.*n),2);
  }

  // higher order corrections
  // E_P = E_n_P_LO * (1 + e_p2<1> + ... + e_P<order>)
  template<int N>
  double e_P(int n, parameters const & p);

  template<>
  double e_P<1>(int n, parameters const & p){
    using Wilson_coefficients::a1_eff;
    using detail::log_n;
    double result = p.contributions.v_Coulomb[0]*(p.alpha_s/(4*pi))*(
	2*a1_eff(p) + 16*beta(0, p.nl)*(log_n(n, p) + harmonic_sum<1>(n+1))
    );
    if(is_threshold_mass(p.mass_scheme)){
      result += 2*detail::delta_m(1, p)/E_n_P_LO(n, p);
    }
    return result;
  }

  //! Leading-order P-wave wave function
  /**
   * @param n    Principal quantum number.
   * @param p    Further parameters.
   * @return     The modulus square of the wave function at the orgin
   */
  inline double F_n_P_LO(int n, parameters const & p){
    return pow(p.m*c_F*p.alpha_s/(2.*n),5)/pi*(n*n-1);
  }

  // higher order corrections
  // F_P = F_n_P_LO * (1 + f_p2<1> + ... + f_P<order>)
  template<int N>
  double f_P(int n, parameters const & p);

  template<>
  double f_P<1>(int n, parameters const & p){
    using Wilson_coefficients::a1_eff;
    using detail::log_n;
    return p.contributions.v_Coulomb[0]*(
	(p.alpha_s/(4*pi)) * (
	    + 5*a1_eff(p)
	    + 4*beta(0, p.nl)*(
		10*log_n(n, p) + (8*(-1 + n + pow(n,2)))/(-1 + pow(n,2)) - (2*n*pi*pi)/3
		- (2*(3 + pow(n,2))*harmonic_sum<1>(n))/(-1 + pow(n,2))
		+ 4*n*harmonic_sum<2>(n)
	    )
	)
    );
  }


  // residue
  template<int N>
  double z_P (int n, parameters const & p);

  template<>
  double z_P<1>(int n, parameters const & p){
    return
      + 2* Wilson_coefficients::ca<1>(p)
      + f_P<1>(n, p)
      ;
  }

  // pole resummation
  static constexpr unsigned P_pole_resummation_max = 3;

  inline complex resum_exp_P(int n, complex lam, parameters const & p){
    if((p.order < 3) || (n == 1)) return 0;
    complex const E = -p.m*pow(c_F*p.alpha_s/(2.*lam),2);
    complex const E_LO = E_n_P_LO(n, p);
    complex const F_LO = F_n_P_LO(n, p);
    complex const e1 = e_P<1>(n, p);
    complex const z1 = z_P<1>(n, p);
    complex const E_N = (1 + e1)*E_LO;
    complex result =
      (e1*E_LO*F_LO)/pow(-E + E_LO,2)
      - F_LO/(-E + E_LO)
      - (F_LO*z1)/(-E + E_LO)
      + (F_LO*(1 + z1))/(-E + E_N)
      ;
    return (
        p.expand_s?
        QCD::n_c/3.*pow(p.m, -4):
        4./3.*QCD::n_c*pow(p.m*(2*p.m + E_N), -2)
    )*result;
  }

  inline complex pole_resummation_P(complex lam, parameters const & p){
    complex result = 0;
    for(unsigned n = 1; n <= P_pole_resummation_max; ++n){
      result += resum_exp_P(n, lam, p);
    }
    return result;
  }

  //! Corrections from insertion-scheme masses to the P-wave Green function
  /**
   * @tparam N      Perturbative order of the Green function before
   *                expansion around the leading-order mass.
   * @tparam DM     Order in expansion around leading-order mass.
   * @param lam     The value of \f$\lambda = \frac{\alpha_s C_F}{2} \sqrt{-\frac{m}{E + i\Gamma}}$\f.
   * @param p       Further parameters.
   * @returns       The threshold-mass correction to the Green function.
   *
   * When using threshold (insertion) mass schemes
   * m_thr = m_Pole + delta_m_0 + delta_m_1 + ...
   * the Green functions should be expanded for small delta_m_1, delta_m2 ...
   *
   * Green_function_P_TM<N, DM> is the expansion term of order DM in the
   * expansion of Green_function_P<N>. Thus, the total perturbative order
   * is N + DM.
   *
   */
  template<int N, int DM>
  complex Green_function_P_TM(complex lam, parameters const & p);

  template<>
  complex Green_function_P_TM<0, 1>(complex lam, parameters const & p){
    using namespace p_wave;
    return
      -c_F*pow(p.m,3)*p.alpha_s/(8*lam*pi)*(
	  -3 + 6*lam + 3*pow(lam,2) + 2*pow(lam,3) - 4*lam*log_lam(lam, p.mu_w, p) -
	  4*lam*harmonic_sum<1>(1 - lam) +2*pow(lam,2)*(-1 + pow(lam,2))*psi<1>(2 - lam)
      )*detail::delta_m(1, p);
  }

}
