/**
 * @file
 * @brief Electron structure and luminosity functions
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.0.0
 *
 * @section LICENSE
 *
 * Copyright 2017 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include "QQbar_threshold/parameters.hpp"

//! -
namespace QQbar_threshold{

  //! Computes the logarithm generated through initial state radiation
  /**
   * @param sqrt_s   center of mass energy
   * @return         \f$\beta = -2*\alpha(m_Z)/\pi [\log(m_e^2/s) + 1]\f$
   *
   * @see alpha_mZ, me, structure_function, luminosity_function, modified_structure_function, modified_luminosity_function
   */
  double top_ISR_log(double sqrt_s);

  //! Computes the logarithm generated through initial state radiation
  /**
   * @param sqrt_s   center of mass energy
   * @return         \f$\beta = -2*\alpha(m_\Upsilon)/\pi [\log(m_e^2/s) + 1]\f$
   *
   * @see alpha_Y, me, structure_function, luminosity_function, modified_structure_function, modified_luminosity_function
   */
  double bottom_ISR_log(double sqrt_s);

  //! Computes the logarithm generated through initial state radiation
  /**
   * @param sqrt_s   center of mass energy
   * @param alpha    electromagnetic coupling constant
   * @return         \f$\beta = -2*\alpha/\pi [\log(m_e^2/s) + 1]\f$
   *
   * @see me, structure_function, luminosity_function, modified_structure_function, modified_luminosity_function
   */
  double ISR_log(double sqrt_s, double alpha);

  //! Computes the electron (or positron) structure function
  /**
   *
   * @param x        fraction of center-of-mass energy squared between 0 and 1
   * @param beta     value of the logarithm to be resummed, see \ref ISR_log
   * @return         the value of the structure function
   *
   * @see me, structure_function, luminosity_function, modified_structure_function, modified_luminosity_function
   */
  double structure_function(double x, double beta);

  //! Modified electron (or positron) structure function
  /**
   *
   * @param t        convolution parameter between 0 and 1
   * @param beta     value of the logarithm to be resummed, see \ref ISR_log
   * @return         the value of the modified structure function
   *
   * The \ref modified_structure_function "modified structure function"
   * \f$\tilde{\Gamma}_{ee}(t, \beta)\f$ is given by
   * \f$\tilde{\Gamma}_{ee}(t, \beta) = 2/\beta t^{2/\beta-1} \Gamma_{ee}(1-t^{2/\beta}, \beta)\f$,
   * where \f$\Gamma_{ee}(1-t^{2/\beta}, \beta)\f$ is the
   * \ref structure_function "electron structure function".
   *
   * @see structure_function, luminosity_function, modified_structure_function, modified_luminosity_function
   */
  double modified_structure_function(double x, double beta);

  //! Computes the electron (or positron) luminosity function
  /**
   *
   * @param x        fraction of center-of-mass energy squared between 0 and 1
   * @param beta     value of the logarithm to be resummed, see \ref ISR_log
   * @return         the value of the luminosity function
   *
   * The luminosity function \f$L\f$ is defined as
   * \f$L(x) = \int_x^1 \frac{dy}{y} f(y)f(x/y)\f$,
   * where \f$f\f$ is the
   * \ref structure_function "electron structure function"
   *
   * @see structure_function, luminosity_function, modified_structure_function, modified_luminosity_function
   */
  double luminosity_function(double x, double beta);

  //! Modified electron (or positron) luminosity function
  /**
   *
   * @param t        convolution parameter between 0 and 1
   * @param beta     value of the logarithm to be resummed, see \ref ISR_log
   * @return         the value of the modified luminosity function
   *
   * The modified luminosity function \f$\tilde{L}(t, \beta)\f$ is defined as
   * \f$\tilde{L}(t, \beta) = t^{1/\beta-1}/\beta L(1-t^{1/\beta}, \beta)\f$,
   * where \f$L\f$ is the
   * \ref luminosity_function "luminosity function"
   *
   * @see structure_function, luminosity_function, modified_structure_function, modified_luminosity_function
   */
  double modified_luminosity_function(double t, double beta);

}
