/**
 * @file
 * @brief Functions to calculate \f$e^+e^- \to Q\bar{Q}\f$ cross sections near threshold
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once
#include <complex>

#include "parameters.hpp"

//! -
namespace QQbar_threshold{
  typedef std::complex<double> complex;

  //! Computes the cross section \f$e^+e^- \to W^+W^-b\bar{b}\f$
  /**
   *
   * @param sqrt_s center-of-mass energy
   * @param mu     renormalisation scales {soft, weak}
   * @param M      top quark mass and width
   * @param o      perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt    further options
   * @return       The cross section in picobarn
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_top}
   *   - options::production = production_channels::all
   *   - options::resum_poles = 6
   *   - options::expand_s = false
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::all
   *   - options::alpha = #alpha_mZ
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 824.12 (for nl == 5)
   *   - options::ml = 0
   *
   * @see options, ttbar_threshold, threshold
   */
  double ttbar_xsection(
      double sqrt_s, scales mu, top_properties M, order o,
      options opt = top_options()
  );
  //! Computes the cross section \f$e^+e^- \to W^+W^-b\bar{b}\f$
  /**
   *
   * @param sqrt_s center-of-mass energy
   * @param mu     overall value for all renormalisation scales
   * @param M      top quark mass and width
   * @param o      perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt    further options
   * @return       The cross section in picobarn
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_top}
   *   - options::production = production_channels::all
   *   - options::resum_poles = 6
   *   - options::expand_s = false
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::all
   *   - options::alpha = #alpha_mZ
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 824.12 (for nl == 5)
   *   - options::ml = 0
   *
   * @see options, ttbar_threshold, threshold
   */
  double ttbar_xsection(
      double sqrt_s, double mu, top_properties M, order o,
      options opt = top_options()
  );
  //! Computes the cross section \f$e^+e^- \to b\bar{b}\f$
  /**
   *
   * @param sqrt_s center-of-mass energy
   * @param mu     soft renormalisation scale
   * @param m      bottom quark mass
   * @param o      perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt    further options
   * @return       The cross section in picobarn
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_bottom}
   *   - options::production = production_channels::all
   *   - options::resum_poles = 6
   *   - options::expand_s = false
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::QED
   *   - options::alpha = #alpha_mZ
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 1220.3 (for nl == 4)
   *   - options::ml = 0
   *
   * @see options, bbbar_threshold, threshold
   */
  double bbbar_xsection(
      double sqrt_s, double mu, double m, order o,
      options opt = bottom_options()
  );

  //! Computes the naive threshold for \f$t\bar{t}\f$ production
  /**
   * @param mu  renormalisation scale
   * @param m   top quark mass
   * @param o   perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt further options
   * @returns The center-of-mass energy of the \f$t\bar{t}\f$ production threshold
   *
   * The naive production threshold is twice the pole quark mass. If the mass is
   * given in a different mass scheme the corresponding pole mass is used.
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_top}
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::all
   *   - options::alpha = #alpha_mZ
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 824.12 (for nl == 5)
   *   - options::ml = 0
   *   - options::double_light_insertion = false
   */
  double ttbar_threshold(
      double mu, double m, order o,
      options opt = top_options()
  );

  //! Computes the naive threshold for \f$b\bar{b}\f$ production
  /**
   * @param mu  renormalisation scale
   * @param m   bottom quark mass
   * @param o   perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt further options
   * @returns The center-of-mass energy of the \f$b\bar{b}\f$ production threshold
   *
   * The naive production threshold is twice the pole quark mass. If the mass is
   * given in a different mass scheme the corresponding pole mass is used.
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_bottom}
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::QED
   *   - options::alpha = #alpha_Y
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 1220.3 (for nl == 4)
   *   - options::ml = 0
   *   - options::double_light_insertion = false
   */
  double bbbar_threshold(
      double mu, double m, order o,
      options opt = bottom_options()
  );

  /** \internal
   * like the usual cross section function, but has a threshold expression
   * as its first argument.
   *
   * @see threshold
   */
  template<typename T>
  inline double ttbar_xsection(
      T thr_expr, scales mu, top_properties M, order o,
      options opt = top_options()
  ){
    return ttbar_xsection(
  	thr_expr.value(ttbar_threshold(mu.soft, M.m, o, opt)),
  	mu, M, o, opt
    );
  }
  template<typename T>
  inline double ttbar_xsection(
      T thr_expr, double mu, top_properties M, order o,
      options opt = top_options()
  ){
    return ttbar_xsection(
  	thr_expr.value(ttbar_threshold(mu, M.m, o, opt)),
  	mu, M, o, opt
    );
  }

  /** \internal
   * like the usual cross section function, but has a threshold expression
   * as its first argument.
   *
   * @see threshold
   */
  template<typename T>
  inline double bbbar_xsection(
      T thr_expr, scales mu, double m, order o,
      options opt = bottom_options()
  ){
    return bbbar_xsection(
  	thr_expr.value(bbbar_threshold(mu.soft, m, o, opt)),
  	mu, m, o, opt
    );
  }
  template<typename T>
  inline double bbbar_xsection(
      T thr_expr, double mu, double m, order o,
      options opt = bottom_options()
  ){
    return bbbar_xsection(
  	thr_expr.value(bbbar_threshold(mu, m, o, opt)),
  	mu, m, o, opt
    );
  }

  //! Bottom contribution to the hadronic R ratio
  /**
   * The bottom contribution to the hadronic R ratio is defined by
   * bbbar_R_ratio = \f$\sigma_b/\sigma_{pt} \f$ with
   * \f$\sigma_{pt} = \frac{4 \pi \alpha(\mu_\alpha)^2}{3s}\f$, where \f$\sigma_b\f$
   * is the bottom production cross section \ref bbbar_xsection
   * converted to natural units. \f$\alpha(\mu_\alpha)\f$ is the running
   * on-shell QED coupling at the scale \f$\mu_\alpha\f$ set
   * by the options.
   *
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_bottom}
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::QED
   *   - options::alpha = #alpha_Y
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 1220.3 (for nl == 4)
   *   - options::ml = 0
   *   - options::double_light_insertion = false
   */
  double bbbar_R_ratio(
      double sqrt_s, double mu, double m, order o,
      options opt = bottom_options()
  );

  //! Top contribution to the hadronic R ratio
  /**
   * The bottom contribution to the hadronic R ratio is defined by
   *  ttbar_R_ratio = \f$\sigma_t/\sigma_{pt} \f$ with
   * \f$\sigma_{pt} = \frac{4 \pi \alpha(\mu_\alpha)^2}{3s}\f$, where \f$\sigma_t\f$
   * is the top production cross section \ref bbbar_xsection
   * converted to natural units. \f$\alpha(\mu_\alpha)\f$ is the running
   * on-shell QED coupling at the scale \f$\mu_\alpha\f$ set
   * by the options.
   *
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_top}
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::all
   *   - options::alpha = #alpha_mZ
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 824.12 (for nl == 5)
   *   - options::ml = 0
   *   - options::double_light_insertion = false
   */
  double ttbar_R_ratio(
      double sqrt_s, scales mu, top_properties M, order o,
      options opt = top_options()
  );

  double ttbar_R_ratio(
      double sqrt_s, double mu, top_properties M, order o,
      options opt = top_options()
  );

  /** \internal
   * like the usual bbbar_R_ratio function, but has a threshold expression
   * as its first argument.
   *
   * @see threshold
   */
  template<typename T>
  inline double bbbar_R_ratio(
      T thr_expr, scales mu, double m, order o,
      options opt = bottom_options()
  ){
    return bbbar_R_ratio(
  	thr_expr.value(bbbar_threshold(mu.soft, m, o, opt)),
  	mu, m, o, opt
    );
  }
  template<typename T>
  inline double bbbar_R_ratio(
      T thr_expr, double mu, double m, order o,
      options opt = bottom_options()
  ){
    return bbbar_R_ratio(
  	thr_expr.value(bbbar_threshold(mu, m, o, opt)),
  	mu, m, o, opt
    );
  }

  /** \internal
   * like the usual ttbar_R_ratio function, but has a threshold expression
   * as its first argument.
   *
   * @see threshold
   */
  template<typename T>
  inline double ttbar_R_ratio(
      T thr_expr, scales mu, top_properties M, order o,
      options opt = top_options()
  ){
    return ttbar_R_ratio(
  	thr_expr.value(ttbar_threshold(mu.soft, M.m, o, opt)),
  	mu, M, o, opt
    );
  }
  template<typename T>
  inline double ttbar_R_ratio(
      T thr_expr, double mu, top_properties M, order o,
      options opt = top_options()
  ){
    return ttbar_R_ratio(
  	thr_expr.value(ttbar_threshold(mu, M.m, o, opt)),
  	mu, M, o, opt
    );
  }

}
