/**
 * @file
 * @brief Implementation of threshold mass correction to Green functions
 *
 * @see Green_functions.hpp
 *
 * TODO: eliminate? (treat like potential insertions into regular Green fun)
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include <complex>
#include <cmath>

#include "complex.hpp"
#include "Green_functions.hpp"
#include "QCD/colour.hpp"
#include "QCD/beta.hpp"
#include "zeta.hpp"
#include "log_lam.hpp"
#include "insertions.hpp"
#include "Wilson_coefficients.hpp"
#include "harmonic_sums.hpp"
#include "hypergeometric.hpp"
#include "scheme_conversion.hpp"

namespace QQbar_threshold{
  namespace detail{
    typedef std::complex<double> complex;

    using namespace complex;
    using namespace zeta;
    using harmonic_sums::harmonic_sum;
    using psi::psi;
    using QCD::beta;
    using QCD::c_F;
    using hypergeometric::PFQ;
    using hypergeometric::der_PFQ;
    using hypergeometric::der2_PFQ;
    using Wilson_coefficients::a1_eff;
    using Wilson_coefficients::a;
    using Wilson_coefficients::b;
    using Wilson_coefficients::b_ep;
    using Wilson_coefficients::v0;
    using Wilson_coefficients::v0_ep;

    /**************************************************************************
                          expansion of LO Green function
    **************************************************************************/

    template<>
    complex Green_function_TM<0,1>(complex lam, parameters const & p){
      return -lam*p.m*delta_m(1, p)/(2*p.alpha_s*c_F*pi)*(
	  1 + 2*lam + 2*pow(lam,2)*psi<1>(1-lam)
      );
    }

    template<>
    complex Green_function_TM<0,2>(complex lam, parameters const & p){
      return  (
	  pow(p.alpha_s*c_F,4)*p.m*delta_m(0, p)*(
	      1 - 2*lam + 2*pow(lam,2)*psi<1>(1-lam)
	  )
	  - 8*pow(lam,2)*(
	      pow(p.alpha_s*c_F,2)*p.m*delta_m(2, p)*(
		  1 + 2*lam + 2*pow(lam,2)*psi<1>(1-lam)
	      )
	      + 2*pow(lam,2)*pow(delta_m(1, p),2)*(
		  -1 - 4*lam
		  - 6*pow(lam,2)*psi<1>(1-lam)
		  + 2*pow(lam,3)*psi<2>(1-lam)
	      )
	  )
      )/(16*pow(p.alpha_s*c_F,3)*lam*pi);
    }

    template<>
    complex Green_function_TM<0,3>(complex lam, parameters const & p){
      return  (
	  - 24*pow(p.alpha_s*c_F,4)*pow(lam,2)*pow(p.m,2)*
	  delta_m(3, p)*(
	      1 + 2*lam + 2*pow(lam,2)*psi<1>(1-lam)
	  )
	  + 3*pow(p.alpha_s*c_F,2)*p.m*delta_m(1, p)*(
	      pow(p.alpha_s*c_F,4)*p.m*(
	      	  1 - 2*lam + 2*pow(lam,2)*psi<1>(1-lam)
	      )
	      - 32*pow(lam,4)*delta_m(2, p)*(
	      	  -1 - 4*lam
	      	  - 6*pow(lam,2)*psi<1>(1-lam)
	      	  + 2*pow(lam,3)*psi<2>(1-lam)
	      )
	      + 4*pow(p.alpha_s*c_F,2)*pow(lam,2)*delta_m(0, p)*(
	      	  1
	      	  - 2*pow(lam,2)*psi<1>(1-lam)
	      	  + 2*pow(lam,3)*psi<2>(1-lam)
	      )
	  )
	  - 64*pow(lam,6)*pow(delta_m(1, p),3)*(
	      3 + 16*lam
	      + 30*pow(lam,2)*psi<1>(1-lam)
	      - 18*pow(lam,3)*psi<2>(1-lam)
	      + 2*pow(lam,4)*psi<3>(1-lam)
	  )
      )/(48*pow(p.alpha_s*c_F,5)*lam*p.m*pi)
	;
    }

    /**************************************************************************
                          expansion of NLO Green function
    **************************************************************************/

    template<>
    complex Green_function_TM<1,1>(complex lam, parameters const & p){
      return p.contributions.v_Coulomb[0]
	*pow(lam,2)*p.m*delta_m(1, p)/(4*c_F*pow(pi,2))*(
	    + a1_eff(p)*(
		-1 - 2*lam*psi<1>(-lam)
		+ pow(lam,2)*psi<2>(-lam)
	    )
	    + 4*beta(0,p.nl)*(
		+ 4*lam*der_PFQ<4>({1,1,1,1},{2,2,1-lam},1)
		- 2*lam*psi<1>(-lam)
		- 2*pow(lam,2)*pow(psi<1>(-lam),2)
		- 2*lam*psi<2>(-lam)
		+ 2*log_lam(lam, p)*(
		    1 + 2*lam*psi<1>(-lam) - pow(lam,2)*psi<2>(-lam)
		)
		- 2*harmonic_sum<1>(-1-lam)*(
		    -1 - 2*lam*psi<1>(-lam) + pow(lam,2)*psi<2>(-lam)
		)
		+ pow(lam,2)*psi<3>(-lam)
	    )
	);
    }

    template<>
    complex Green_function_TM<1,2>(complex lam, parameters const & p){
      return p.contributions.v_Coulomb[0]*(
	  + pow(p.alpha_s,2)*pow(c_F,4)*p.m*delta_m(0, p)*(
	      - a1_eff(p)*(
	  	  1 - 2*lam*psi<1>(-lam) + pow(lam,2)*psi<2>(-lam)
	      )
	      + 4*beta(0,p.nl)*(
	  	  + 2*gamma_Euler
	  	  - 4*lam*der_PFQ<4>({1,1,1,1},{2,2,1-lam},1)
	  	  - 2*lam*psi<1>(-lam)
	  	  - 4*gamma_Euler*lam*psi<1>(-lam)
	  	  + 2*pow(lam,2)*pow(psi<1>(-lam),2)
	  	  + 2*lam*psi<2>(-lam)
	  	  + 2*gamma_Euler*pow(lam,2)*psi<2>(-lam)
	  	  + 2*log_lam(lam, p)*(
	  	      + 1 - 2*lam*psi<1>(-lam)
	  	      + pow(lam,2)*psi<2>(-lam)
	  	  )
	  	  + 2*psi<0>(-lam)*(
	  	      + 1 - 2*lam*psi<1>(-lam)
	  	      + pow(lam,2)*psi<2>(-lam)
	  	  )
	  	  - pow(lam,2)*psi<3>(-lam)
	      )
	  )
	  + 8*pow(c_F,2)*pow(lam,2)*p.m*delta_m(2, p)*(
	      + a1_eff(p)*(
	  	  -1 - 2*lam*psi<1>(-lam) + pow(lam,2)*psi<2>(-lam)
	      )
	      + 4*beta(0,p.nl)*(
	  	  + 2*gamma_Euler
	  	  + 4*lam*der_PFQ<4>({1,1,1,1},{2,2,1-lam},1)
	  	  - 2*lam*psi<1>(-lam)
	  	  + 4*gamma_Euler*lam*psi<1>(-lam)
	  	  - 2*pow(lam,2)*pow(psi<1>(-lam),2)
	  	  - 2*lam*psi<2>(-lam)
	  	  - 2*gamma_Euler*pow(lam,2)*psi<2>(-lam)
	  	  + 2*log_lam(lam, p)*(
	  	      1 + 2*lam*psi<1>(-lam)
	  	      - pow(lam,2)*psi<2>(-lam)
	  	  )
	  	  - 2*psi<0>(-lam)*(
	  	      -1 - 2*lam*psi<1>(-lam) + pow(lam,2)*psi<2>(-lam)
	  	  )
	  	  + pow(lam,2)*psi<3>(-lam)
	      )
	  )
	  + (
	      + 16*pow(lam,4)*pow(delta_m(1, p),2)*(
	  	  a1_eff(p)*(
	  	      2 + 6*lam*psi<1>(-lam)
	  	      - 6*pow(lam,2)*psi<2>(-lam)
	  	      + pow(lam,3)*psi<3>(-lam)
	  	  )
	  	  + 4*beta(0,p.nl)*(
	  	      2 - 4*gamma_Euler
	  	      - 12*lam*der_PFQ<4>({1,1,1,1},{2,2,1-lam},1)
	  	      + 4*pow(lam,2)*der2_PFQ<4>({1,1,1,1},{2,2,1-lam},1)
	  	      - 4*psi<0>(-lam)
	  	      + 12*lam*psi<1>(-lam)
	  	      - 12*gamma_Euler*lam*psi<1>(-lam)
	  	      - 12*lam*psi<0>(-lam)*psi<1>(-lam)
	  	      + 12*pow(lam,2)*pow(psi<1>(-lam),2)
	  	      + 6*lam*psi<2>(-lam)
	  	      - 4*pow(lam,2)*psi<2>(-lam)
	  	      + 12*gamma_Euler*pow(lam,2)*psi<2>(-lam)
	  	      + 12*pow(lam,2)*psi<0>(-lam)*psi<2>(-lam)
	  	      - 6*pow(lam,3)*psi<1>(-lam)*psi<2>(-lam)
	  	      - 6*pow(lam,2)*psi<3>(-lam)
	  	      - 2*gamma_Euler*pow(lam,3)*psi<3>(-lam)
	  	      - 2*pow(lam,3)*psi<0>(-lam)*psi<3>(-lam)
	  	      - 2*log_lam(lam, p)*(
	  		  2 + 6*lam*psi<1>(-lam)
	  		  -6*pow(lam,2)*psi<2>(-lam)
	  		  + pow(lam,3)*psi<3>(-lam)
	  	      )
	  	      + pow(lam,3)*psi<4>(-lam)
	  	  )
	      )
	  )/pow(p.alpha_s,2)
      )/(32*pow(c_F,3)*pow(pi,2));
    }

    /**************************************************************************
                         expansion of NNLO Green function
    **************************************************************************/

    template<>
    complex Green_function_TM<2,1>(complex lam, parameters const & p){
      return
	+ pow(p.contributions.v_Coulomb[0], 2)*(
	    p.alpha_s*lam*p.m*delta_m(1, p)*(
		- 24*a1_eff(p)*4*beta(0,p.nl)*pow(lam,3)
		*der_insertion_sum1<1, 1+ep>( lam)
		+ 24*pow(4*beta(0,p.nl),2)*pow(lam,3)
		*der_insertion_sum1<1+ep, 1+ep>( lam)
		- 24*pow(4*beta(0,p.nl),2)*pow(lam,2)
		*insertion_sum1<1+ep, 1+ep>( lam)
		- 24*4*beta(0,p.nl)*pow(lam,2)
		*insertion_sum1<1, 1+ep>( lam)*(
		    -a1_eff(p) + 2*4*beta(0,p.nl)*(-1 + log_lam(lam, p))
		)
		+ 48*pow(4*beta(0,p.nl),2)*pow(lam,3)
		*der_insertion_sum1<1, 1+ep>(lam)
		*log_lam(lam, p)
		+ 24*pow(4*beta(0,p.nl),2)*psi<0>(-lam)*(-2 + pow(lam,3)*psi<2>(-lam))
		- pow(lam,2)*(
		    6*(
			pow(a1_eff(p),2) + 4*a1_eff(p)*4*beta(0,p.nl) + 8*pow(4*beta(0,p.nl),2)*gamma_Euler
		    )*psi<1>(1-lam)
		    - 48*pow(4*beta(0,p.nl),2)*gamma_Euler*psi<1>(-lam)
		    - 12*pow(a1_eff(p),2)*lam*psi<2>(1-lam)
		    - 12*a1_eff(p)*4*beta(0,p.nl)*lam*psi<2>(1-lam)
		    - 24*pow(4*beta(0,p.nl),2)*gamma_Euler*lam*psi<2>(1-lam)
		    + 24*pow(4*beta(0,p.nl),2)*lam*psi<0>(1-lam)*psi<2>(1-lam)
		    + 24*pow(4*beta(0,p.nl),2)*psi<2>(-lam)
		    + 24*pow(4*beta(0,p.nl),2)*gamma_Euler*lam*psi<2>(-lam)
		    + 8*pow(4*beta(0,p.nl),2)*lam*psi<3>(1-lam)
		    + 3*pow(a1_eff(p),2)*pow(lam,2)*psi<3>(1-lam)
		    - 8*pow(4*beta(0,p.nl),2)*lam*psi<3>(-lam)
		)
		- 12*pow(4*beta(0,p.nl),2)*pow(lam,2)*pow(log_lam(lam, p),2)*(
		    2*psi<1>(-lam) + lam*(-4*psi<2>(-lam) +lam*psi<3>(-lam))
		)
		+ 4*4*beta(0,p.nl)*log_lam(lam, p)*(
		    + 6*4*beta(0,p.nl)*psi<0>(-lam)*(
			-2 - 2*pow(lam,3)*psi<2>(-lam) +pow(lam,4)*psi<3>(-lam)
		    )
		    + lam*(
			+ 6*lam*psi<1>(1-lam)*(
			    a1_eff(p) + 2*4*beta(0,p.nl)*gamma_Euler
			    - 4*beta(0,p.nl)*pow(lam,2)*psi<2>(1-lam)
			)
			+ 6*4*beta(0,p.nl)*psi<1>(-lam)*(
			    - 2
			    - 2*(-1 + gamma_Euler)*lam
			    + pow(lam,3)*psi<2>(-lam)
			)
			+ lam*(
			    -12*lam*(
				a1_eff(p)
				+ 2*4*beta(0,p.nl)*gamma_Euler
				- 4*beta(0,p.nl)*psi<0>(1-lam)
			    )*psi<2>(1-lam)
			    + 6*4*beta(0,p.nl)*(
				1
				+ (-1 + 4*gamma_Euler)*lam
			    )*psi<2>(-lam)
			    + lam*(
				(
				    + 4*4*beta(0,p.nl)
				    + 3*a1_eff(p)*lam
				    + 6*4*beta(0,p.nl)*gamma_Euler*lam
				    - 6*4*beta(0,p.nl)*lam*psi<0>(1-lam)
				)*psi<3>(1-lam)
				- 2*4*beta(0,p.nl)*(
				    + (5 + 3*gamma_Euler*lam)*psi<3>(-lam)
				    - 24/pow(lam,4)
				)
			    )
			)
		    )
		)
	    )
	)/(96*c_F*pow(pi,3))
	+ p.contributions.v_delta[0]*(
	    p.alpha_s/(4*pi)*c_F*p.m*delta_m(1, p)*(
		2*v0(p.nl) - 2*lam*v0(p.nl) +4*gamma_Euler*lam*v0(p.nl)
		- lam*v0_ep(p.nl)
		+4*lam*v0(p.nl)*log_lam(lam, p)
		+4*lam*v0(p.nl)*psi<0>(1-lam)
	    )*(1 + 2*lam +2*pow(lam,2)*psi<1>(1-lam))
	)
	+ p.contributions.v_kinetic[0]*(
	    p.alpha_s/(64*pi)*c_F*p.m*delta_m(1, p)*(
		+ 16*log_lam(lam, p)*(
		    2 + lam + 2*pow(lam,2)
		    + 2*pow(lam,3)*psi<1>(1-lam)
		)
		+ (
		    15 + 4*lam + 32*gamma_Euler*lam
		    - 8*pow(lam,2) +16*gamma_Euler*pow(lam,2)
		    - 48*pow(lam,3) + 32*gamma_Euler*pow(lam,3)
		    + 16*lam*psi<0>(-lam)
		    + 10*pow(lam,2)*psi<1>(1-lam)
		    +16*pow(lam,3)*psi<1>(1-lam)
		    -48*pow(lam,4)*psi<1>(1-lam)
		    + 32*gamma_Euler*pow(lam,4)*psi<1>(1-lam)
		    + 16*lam*psi<0>(1-lam)*(
			1 + lam + 2*pow(lam,2) + 2*pow(lam,3)*psi<1>(1-lam)
		    )
		    +2*pow(lam,3)*psi<2>(1-lam)
		    - 8*pow(lam,3)*psi<2>(-lam)
		)/lam
	    )
	)
	+ p.contributions.v_r2inv[0]*(
	    p.alpha_s/(8*pi)*lam*p.m*delta_m(1, p)*(
		+ 2*b<1>(p.nl) + b_ep<1>(p.nl) - 4*b<1>(p.nl)*gamma_Euler
		+ 12*b<1>(p.nl)*lam + 2*b_ep<1>(p.nl)*lam
		- 8*b<1>(p.nl)*gamma_Euler*lam
		- b<1>(p.nl)*log(4) - b<1>(p.nl)*lam*log(16)
		- lam*(
		    -2*b_ep<1>(p.nl)*lam + b<1>(p.nl)*(
			4 + lam*(-12 + 8*gamma_Euler + log(16))
		    )
		)*psi<1>(1-lam)
		- 4*b<1>(p.nl)*log_lam(lam, p)*(
		    1 + 2*lam + 2*pow(lam,2)*psi<1>(1-lam)
		)
		- 4*b<1>(p.nl)*psi<0>(1-lam)*(
		    1 + 2*lam + 2*pow(lam,2)*psi<1>(1-lam)
		)
	    )
	)
	+ p.contributions.v_p2[0]*(
	    p.alpha_s/(4*pi)*c_F*p.m*delta_m(1, p)*(
		2*gamma_Euler - 2*lam + 4*gamma_Euler*lam
		- 12*pow(lam,2) + 8*gamma_Euler*pow(lam,2)
		+2*psi<0>(-lam) + 4*pow(lam,2)*psi<1>(1-lam)
		-12*pow(lam,3)*psi<1>(1-lam)
		+ 8*gamma_Euler*pow(lam,3)*psi<1>(1-lam)
		+ 4*lam*psi<0>(1-lam)*(
		    1 + 2*lam + 2*pow(lam,2)*psi<1>(1-lam)
		)
		+2*log_lam(lam, p)*(
		    1 + 2*lam + 4*pow(lam,2) +4*pow(lam,3)*psi<1>(1-lam)
		)
		- pow(lam,2)*psi<2>(-lam)
	    )
	)
	+ p.contributions.v_Coulomb[1]*(
	    p.alpha_s/(48*c_F*pow(pi,3))*lam*p.m*delta_m(1, p)*(
		- 48*pow(4*beta(0,p.nl),2)
		+ 48*pow(4*beta(0,p.nl),2)*gamma_Euler
		- 3*a<2>(p.nl)*lam
		+ 12*a<1>(p.nl)*4*beta(0,p.nl)*gamma_Euler*lam
		+ 6*16*beta(1, p.nl)*gamma_Euler*lam
		- 12*pow(4*beta(0,p.nl),2)*pow(gamma_Euler,2)*lam
		- pow(4*beta(0,p.nl),2)*lam*pow(pi,2)
		- 96*pow(4*beta(0,p.nl),2)*pow(lam,2)*der_PFQ<5>({1,1,1,1,1},{2,2,2,1-lam},1)
		+ 12*pow(4*beta(0,p.nl),2)*pow(lam,2)
		*der_insertion_sum1<1+2*ep>( lam)
		- 12*lam*der_PFQ<4>({1,1,1,1},{2,2,1-lam},1)*(
		    -2*pow(4*beta(0,p.nl),2)
		    - 2*a<1>(p.nl)*4*beta(0,p.nl)*lam
		    + 4*pow(4*beta(0,p.nl),2)*lam
		    - 16*beta(1, p.nl)*lam
		    + 4*pow(4*beta(0,p.nl),2)*lam*log_lam(lam, p)
		    +2*pow(4*beta(0,p.nl),2)*lam*psi<0>(-lam)
		)
		+ 6*pow(lam,2)*(
		    - 2*a<1>(p.nl)*4*beta(0,p.nl)*lam
		    - 16*beta(1, p.nl)*lam
		    + 4*pow(4*beta(0,p.nl),2)*(
			-2 + (-1 + gamma_Euler)*lam
		    )
		)*pow(psi<1>(-lam),2)
		- 24*pow(4*beta(0,p.nl),2)*PFQ<4>({1, 1, 1, 1}, {2, 2, 1 - lam}, 1)*(
		    -1 + 2*lam + pow(lam,2)*psi<1>(-lam)
		)
		+ psi<2>(-lam)*(
		    + pow(lam,2)*(
			- 12*a<1>(p.nl)*4*beta(0,p.nl)
			- 12*pow(4*beta(0,p.nl),2)
			- 6*16*beta(1, p.nl)
			- 48*pow(4*beta(0,p.nl),2)*gamma_Euler
		    )
		    + pow(lam,3)*(
			+ 3*a<2>(p.nl)
			+ 96*pow(4*beta(0,p.nl),2)
			- 12*a<1>(p.nl)*4*beta(0,p.nl)*gamma_Euler
			- 48*pow(4*beta(0,p.nl),2)*gamma_Euler
			- 6*16*beta(1, p.nl)*gamma_Euler
			+ 12*pow(4*beta(0,p.nl),2)*pow(gamma_Euler,2)
			+ 9*pow(4*beta(0,p.nl),2)*pow(pi,2)
		    )
		)
		+ 12*pow(4*beta(0,p.nl),2)*lam*pow(log_lam(lam, p),2)*(
		    -1 - 2*lam*psi<1>(-lam) + pow(lam,2)*psi<2>(-lam)
		)
		+ 12*pow(4*beta(0,p.nl),2)*lam*pow(psi<0>(-lam),2)*(
		    -1 - 2*lam*psi<1>(-lam) + pow(lam,2)*psi<2>(-lam)
		)
		+ 28*pow(4*beta(0,p.nl),2)*pow(lam,2)*psi<3>(-lam)
		+ 6*a<1>(p.nl)*4*beta(0,p.nl)*pow(lam,3)*psi<3>(-lam)
		+ 3*16*beta(1, p.nl)*pow(lam,3)*psi<3>(-lam)
		+12*pow(4*beta(0,p.nl),2)*gamma_Euler*pow(lam,3)*psi<3>(-lam)
		+6*lam*log_lam(lam, p)*(
		    2*a<1>(p.nl)*4*beta(0,p.nl) + 16*beta(1, p.nl) - 4*pow(4*beta(0,p.nl),2)*gamma_Euler
		    +2*(
			2*a<1>(p.nl)*4*beta(0,p.nl) + 16*beta(1, p.nl) + pow(4*beta(0,p.nl),2)*(2 - 4*gamma_Euler)
		    )*lam*psi<1>(-lam)
		    + 4*pow(4*beta(0,p.nl),2)*pow(lam,2)*pow(psi<1>(-lam),2)
		    + 4*pow(4*beta(0,p.nl),2)*lam*psi<2>(-lam)
		    - 2*a<1>(p.nl)*4*beta(0,p.nl)*pow(lam,2)*psi<2>(-lam)
		    - 16*beta(1, p.nl)*pow(lam,2)*psi<2>(-lam)
		    + 4*pow(4*beta(0,p.nl),2)*gamma_Euler*pow(lam,2)*psi<2>(-lam)
		    + 4*pow(4*beta(0,p.nl),2)*psi<0>(-lam)*(
			-1 - 2*lam*psi<1>(-lam) +pow(lam,2)*psi<2>(-lam)
		    )
		    - 2*pow(4*beta(0,p.nl),2)*pow(lam,2)*psi<3>(-lam)
		)
		- 8*pow(4*beta(0,p.nl),2)*pow(lam,3)*psi<4>(-lam)
		+ 6*psi<0>(-lam)*(
		    + 4*pow(4*beta(0,p.nl),2) + 2*a<1>(p.nl)*4*beta(0,p.nl)*lam + 16*beta(1, p.nl)*lam
		    - 4*pow(4*beta(0,p.nl),2)*gamma_Euler*lam
		    + 2*(
			2*a<1>(p.nl)*4*beta(0,p.nl) + 16*beta(1, p.nl) -4*pow(4*beta(0,p.nl),2)*(-1 + gamma_Euler)
		    )*pow(lam,2)*psi<1>(-lam)
		    + 4*pow(4*beta(0,p.nl),2)*pow(lam,3)*pow(psi<1>(-lam),2)
		    + pow(lam,2)*(
			-2*a<1>(p.nl)*4*beta(0,p.nl)*lam - 16*beta(1, p.nl)*lam
			+ 4*pow(4*beta(0,p.nl),2)*(-2 + (-1 + gamma_Euler)*lam)
		    )*psi<2>(-lam)
		    + 2*pow(4*beta(0,p.nl),2)*pow(lam,3)*psi<3>(-lam)
		    - 8*pow(4*beta(0,p.nl),2)*pow(lam,2)*zeta_3
		)
		+ 6*lam*psi<1>(-lam)*(
		    2*a<1>(p.nl)*4*beta(0,p.nl)*(-1 + 2*gamma_Euler)*lam
		    - (a<2>(p.nl) + 16*beta(1, p.nl) - 2*16*beta(1, p.nl)*gamma_Euler)*lam
		    + 6*pow(4*beta(0,p.nl),2)*pow(lam,2)*psi<2>(-lam)
		    + pow(4*beta(0,p.nl),2)*(
			+ 6
			- lam*(24 - 12*gamma_Euler + 4*pow(gamma_Euler,2) +3*pow(pi,2))
			+ 8*pow(lam,2)*zeta_3)
		)
	    )
	)
	;
    }

  }
}
