/**
 * @file
 * @brief Electroweak contribution to current matching coefficient
 *
 * Note that corrections including weak (W, Z) bosons are discarded
 * if the heavy quark is too light, i.e. m < mZ
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.0.0
 * @section LICENSE
 *
 * Copyright 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include <cmath>
#include "Wilson_coefficients.hpp"
#include "zeta.hpp"
#include "dilog.hpp"
#include "complex.hpp"

namespace QQbar_threshold{

  namespace Wilson_coefficients{

    namespace {
      using zeta::pi;
      using zeta::zeta_2;
      using dilog::Li2;
      using std::abs;
      using std::log;

      using detail::coupling;
      using detail::vector;
      using detail::axial;
      using detail::c_e;
      using detail::c_nu;
      const double v_e = c_e<vector>();
      const double a_e = c_e<axial>();
      const double v_nu = c_nu<vector>();
      const double a_nu = c_nu<axial>();
      const double v_u = vector_Z_coupling(T3_u, e_u);
      const double a_u = axial_Z_coupling(T3_u, e_u);
      const double v_d = vector_Z_coupling(T3_d, e_d);
      const double a_d = axial_Z_coupling(T3_d, e_d);

      double Z_prop(double s){
        return 1/(s - mZ*mZ);
      }

      /*********** master integrals **********/
      /**
       * formulas from Appendix A of \ref Guth:1991ab
       * or http://qcdloop.fnal.gov/
       * always drop the imaginary part
       */

      namespace {

        // @see Eq. (A.2) in \ref Guth:1991ab
        template<int i>
        double l(double r);

        template<>
        double l<1>(double r){
          return (1 - r)*log(abs(1/r - 1));
        }

        template<>
        double l<2>(double r){
          return log(4/r);
        }

        template<>
        double l<3>(double r){
          assert(r < 1);
          const double beta = sqrt(1 - r);
          return beta*log((1 + beta)/(1 - beta));
        }

        template<>
        double l<4>(double r){
          if(r <= 4) return sqrt(r*(4-r))*acos(sqrt(r)/2);
          const double root = sqrt(1 - 4/r);
          return - r*root/2*log((1 + root)/(1 - root));
        }

        double l5(double rh, double rz){
          const double radicand_plus = pow(sqrt(rh) + sqrt(rz), 2) - 4;
          const double sqrt_plus = sqrt(abs(radicand_plus));
          const double radicand_minus = 4 - pow(sqrt(rh) - sqrt(rz), 2);
          const double sqrt_minus = sqrt(abs(radicand_minus));
          const double fact = 1./4.*sqrt_plus*sqrt_minus;
          if(radicand_minus < 0){
            return fact*log(
                (sqrt_plus + sqrt_minus)/abs(sqrt_plus - sqrt_minus)
            );
          }
          if(radicand_plus < 0){
            return -fact*log(
                (sqrt_plus + sqrt_minus)/abs(sqrt_plus - sqrt_minus)
            );
          }
          return -2*fact*atan(sqrt_minus/sqrt_plus);
        }

        //! A_0 function; made dimensionless
        double A_0(double m){
          if(m == 0) return 0;
          return 1 - 2*log(m);
        }

        double A(double m1, double m2){
          assert(m1 <= m2);
          if(m1 == 0) return A_0(m2);
          if(m1 == m2) return - 2*log(m1);
          return 1/(m1*m1 - m2*m2)*(m1*m1*A_0(m1) - m2*m2*A_0(m2));
        }

        double B_0(double s, double m1, double m2){
          using namespace complex;
          static constexpr double ep = 1e-15;
          if(m1 > m2) std::swap(m1, m2);
          if(s == 0) return A(m1, m2);
          if(m1 == 0){
            if(m2 == 0) return 2 - log(s);
            const double r = m2*m2/s;
            if(r == 1) return 2 - 2*log(m2);
            return 2 - 2*log(m2) - l<1>(r);
          }
          assert(m2 != 0);
          const double del = s - m2*m2 + m1*m1;
          complex const disc = del*del - 4*s*m1*m1*complex{1, -ep};
          std::array<complex , 2> lam = {
            (del + sqrt(disc))/(2*s),
            (del - sqrt(disc))/(2*s)
          };
          double result = 2 - log(s);
          for(auto lam_i: lam){
            result += real(lam_i*log((lam_i - 1)/lam_i) - log(lam_i - 1));
          }
          return result;
        }
      }

      //! Electroweak box corrections
      /**
       * @see Eq. (19) in \ref Hoang:2006pd
       * The overall factor alpha/(4*pi) has been
       * drawn out of the Wilson coefficient
       */

       namespace {

        template<int i>
        double f(double r);

        // @see Eq. (35) in \ref Hoang:2006pd
        template<>
        double f<1>(double r){
          assert((r > 0) && (r < 1));
          const double beta = sqrt(1 - r);
          return pi*pi - pow(log((1+beta)/(1-beta)), 2);
        }

        // @see Eq. (A.3) in \ref Guth:1991ab
        template<>
        double f<2>(double r){
          static constexpr double ep = 1e-15;
          return 3*zeta_2 + 2*(real(Li2(2/r + complex{-1, ep})) - Li2(1 - 2/r));
        }

        // @see Eq. (A.3) in \ref Guth:1991ab
        template<>
        double f<3>(double r){
          return zeta_2 - Li2(1 - 4/r);
        }

        // @see Eq. (34) in \ref Hoang:2006pd
        double FWW(double r){
          return 2/(r-1)*(l<1>(r) - l<3>(r) + 1./4.*(f<2>(r) - r*f<1>(r)));
        }
        // @see Eq. (11) in \ref Guth:1991ab
        double FZZ(double r){
          return 1/(2 - r)*(
              + (2-r)*log(4*r) - 4*l<4>(r) + (2 + r)*l<3>(r)
              + (-2 + r*(1 - r/4))*f<3>(r) + r*(1 - r/4)*f<1>(r)
          );
        }
        // @see Eq. (36) in \ref Hoang:2006pd
        double FZA(double r){
          assert(r < 4);
          return
            2*log(r) + log(4/r) + (1 + r/4)*log(4/r - 1)
            - 4/(4 - r)*l<4>(r) + (r/4 - 1)*f<3>(r)
            ;
        }

        /**
         * @see Eq. (3.13) in \ref Grzadkowski:1986pm
         *      Eq. (10) in \ref Guth:1991ab
         *
         */

        template<coupling c>
        double hEW_WW(parameters const & p){
          return -1/(4*sw2*sw2)*FWW(mW*mW/(p.m*p.m));
        }

        template<coupling c>
        double hEW_ZZ(parameters const & p);
        template<>
        double hEW_ZZ<vector>(parameters const & p){
          const double a_q = p.c_q[axial];
          const double v_q = p.c_q[vector];
          return -8*a_e*v_e*a_q*v_q*FZZ(mZ*mZ/(p.m*p.m));
        }
        template<>
        double hEW_ZZ<axial>(parameters const & p){
          const double a_q = p.c_q[axial];
          const double v_q = p.c_q[vector];
          return -4*(a_e*a_e + v_e*v_e)*a_q*v_q*FZZ(mZ*mZ/(p.m*p.m));
        }

        template<coupling c>
        double hEW_ZA(parameters const & p);
        template<>
        double hEW_ZA<vector>(parameters const & p){
          const double a_q = p.c_q[axial];
          return -4*a_e*e*a_q*p.e_q*FZA(mZ*mZ/(p.m*p.m));
        }
        template<>
        double hEW_ZA<axial>(parameters const & p){
          const double a_q = p.c_q[axial];
          return -4*v_e*e*a_q*p.e_q*FZA(mZ*mZ/(p.m*p.m));
        }

        template<coupling c>
        double hEW_box(parameters const & p){
          if(p.production == production_channel::photon_only) return 0.;
          /*
           * for the total cross section we have drawn out a factor of 1/s,
           * which is not present in the box corrections
           */
          const double s_cancel = p.s/(4*p.m*p.m);
          return s_cancel*(hEW_WW<c>(p) + hEW_ZZ<c>(p) + hEW_ZA<c>(p));
        }

      }

      //! Electroweak corrections to quark current
      /**
       * @see Eq. (18) in \ref Hoang:2006pd
       * The overall factor alpha/(4*pi) has been
       * drawn out of the Wilson coefficient.
       * Note that the normalisation of sum_a functions
       * differs by a factor sw2 to be compatible with
       * \ref Guth:1991ab
       *  The split into particle-dependent coefficients
       *  and remainder is slightly different here!
       */
      template<coupling C>
      double hEW_QQbar(parameters const & p);

      namespace {

        enum class production_particle{
          A, Z
        };

        /** a functions
         *
         *  Note that the normalisation of the a functions
         *  differs by a factor 1/sw2 from \ref Hoang:2006pd.
         */

        // @see Eq. (13) in \ref Guth:1991ab
        template<production_particle P>
        std::array<double, 3>  a_W_coeffs(parameters const & p);
        template<>
        std::array<double, 3>  a_W_coeffs<production_particle::A>(
            parameters const & p
        ){
          return {{p.e_q, sqrt(sw2*cw2), 0}};
        }
        template<>
        std::array<double, 3>  a_W_coeffs<production_particle::Z>(
            parameters const & p
        ){
          const double v_q = p.c_q[vector];
          const double a_q = p.c_q[axial];
          return {{a_q + v_q , cw2, a_q}};
        }
        template<production_particle P>
        double a_W(double r, parameters const & p){
          const double a_q = p.c_q[axial];
          const std::array<double, 3> c = a_W_coeffs<P>(p);
          return
            + c[0]*1./3.*(
                1 - 2*r + 2*(r + 1/(1+r))*l<1>(r) - 2/(1 + r)*l<2>(r)
            )
            + c[1]*4./3.*a_q*(
                4 - (2/(1 + r) - 1)*l<1>(r) + 2/(1 + r)*l<2>(r) - 5./2.*l<3>(r)
            )
            + c[2]*(-1./2. + r*(1 - l<1>(r)))
            ;
        }

        // @see Eq. (14) in \ref Guth:1991ab
        template<production_particle P>
        double a_phi_W(double r, parameters const & p){
          const double a_q = p.c_q[axial];
          const std::array<double, 3> c = a_W_coeffs<P>(p);
          return (
              + c[0]*1./3.*(
                  1 - 2*r + 2*(r + 1/(1+r))*l<1>(r) - 2/(1 + r)*l<2>(r)
              )
              + c[1]*4./3.*a_q*(
                  1 - (2/(1 + r) - 1)*l<1>(r) + 2/(1 + r)*l<2>(r) - l<3>(r)
              )
              + c[2]*1./3.*(-5./2. + r - (r + 2)*l<1>(r) + 2*l<3>(r))
          )/(2*r);
        }

        // @see Eq. (15) in \ref Guth:1991ab
        template<production_particle P>
        std::array<double, 2>  a_Z_coeffs(parameters const & p);
        template<>
        std::array<double, 2>  a_Z_coeffs<production_particle::A>(parameters const & p){
          const double v_q = p.c_q[vector];
          const double a_q = p.c_q[axial];
          return {{p.e_q*(a_q*a_q + v_q*v_q), p.e_q}};
        }
        template<>
        std::array<double, 2>  a_Z_coeffs<production_particle::Z>(parameters const & p){
          const double v_q = p.c_q[vector];
          const double a_q = p.c_q[axial];
          return {{v_q*(3*a_q*a_q +v_q*v_q), v_q}};
        }
        template<production_particle P>
        double a_Z(double r, parameters const & p){
          const double a_q = p.c_q[axial];
          const double v_q = p.c_q[vector];
          const std::array<double, 2> c = a_Z_coeffs<P>(p);
          return
            + c[0]*2./3.*sw2*(
                -3 + 2*r + (3 - r)*r*log(r) + 2*(1 - r)*l<4>(r)
            )
            + c[1]*sw2*(
                2*(a_q*a_q + v_q*v_q)*(
                    + 3 - 2*r
                    + (2 + r*(r - 3))*log(r)
                    + 2/r*(4/(r - 4) + 2 + r*(r - 1))*l<4>(r)
                )
                + 8*(v_q*v_q - a_q*a_q)*(
                    -1 + (r - 1)/2*log(r) + (1 - 1/(4 - r))*l<4>(r)
                )
            );
        }

        // @see Eq. (16) in \ref Guth:1991ab
        template<production_particle P>
        double a_phi_Z_coeff(parameters const & p);
        template<>
        double a_phi_Z_coeff<production_particle::A>(parameters const & p){
          return p.e_q;
        }
        template<>
        double a_phi_Z_coeff<production_particle::Z>(parameters const & p){
          return p.c_q[vector];
        }
        template<production_particle P>
        double a_phi_Z(double r, parameters const & p){
          return a_phi_Z_coeff<P>(p)/(12*cw2*r)*(
              -4*r - 2*(6/(4 - r) - 1 -2*r)*l<4>(r) - (3 - 2*r)*r*log(r)
          );
        }

        // @see Eq. (37) in \ref Hoang:2006pd
        template<production_particle P>
        double a_ZH(double rH, double rZ, parameters const & p);

        template<>
        double a_ZH<production_particle::A>(
            double, double, parameters const &
        ){
          return 0;
        }

        template<>
        double a_ZH<production_particle::Z>(
            double rH, double rZ, parameters const & p
        ){
          const double v_q = p.c_q[vector];
          const double Yukawa_factor = sqrt(
              p.Yukawa_t2*sw2*cw2/(2*pi*p.alpha)
          )*mZ/p.m;
          return Yukawa_factor*v_q/(2*cw2)*(
              + (rH + rZ)/8*log(rH/rZ)
              + (
                  - (4 - rH + rZ)*l5(rH, rZ)
                  + (4 - rH)*((1 - rH/2)/2*log(rH*rZ) - l<4>(rH))
                  + rZ*((1 - rZ/2)/2*log(rH*rZ) - l<4>(rZ))
              )/(rZ + rH - 4)
          );
        }

        //! Sum over a functions
        /**
         * @see Eq. (12) in \ref Guth:1991ab
         *
         * a_H is already contained in the Higgs corrections
         */
        template<production_particle P>
        double sum_a(parameters const & p){
          const double rW = mW*mW/(p.m*p.m);
          const double rZ = mZ*mZ/(p.m*p.m);
          const double rH = p.m_Higgs*p.m_Higgs/(p.m*p.m);
          return
            + a_W<P>(rW, p) + a_phi_W<P>(rW, p)
            + a_Z<P>(rZ, p) + a_phi_Z<P>(rZ, p)
            + a_ZH<P>(rH, rZ, p)
            ;
        }

        template<coupling c>
        double hEW_QQbar_Zsum(parameters const & p){
          return c_e<c>()*p.s*Z_prop(p.s)*sum_a<production_particle::Z>(p);
        }

      }

      template<>
      double hEW_QQbar<axial>(parameters const & p){
        assert(p.production != production_channel::photon_only);
        return  hEW_QQbar_Zsum<axial>(p)/sw2;
      }

      template<>
      double hEW_QQbar<vector>(parameters const & p){
        double photon_contr = e*sum_a<production_particle::A>(p);
        if(p.production == production_channel::photon_only){
          return photon_contr/sw2;
        }
        return (photon_contr + hEW_QQbar_Zsum<vector>(p))/sw2;
      }

      //! pure QED correction to QQbar vertex
      template<int N>
      double cv_QED_QQbar(parameters const & p);

      template<>
      double cv_QED_QQbar<2>(parameters const &){
        return -8;
      }

      //! Electroweak corrections to leptonic current
      /**
       * @see Eq. (17) in \ref Hoang:2006pd
       */
      template<coupling c>
      double hEW_ee(parameters const & p);

      namespace {

        // @see Eqs. (B.3) in \ref Grzadkowski:1986pm
        double rho(double u){
          static constexpr double ep = 1e-15;
          return
            + 1./2. - log(u) + 4*(1 + 1/(2*u))*(log(u) - 1)
            - 2*pow(1 + 1/u, 2)*(zeta_2 - real(Li2(1. + u*complex{1,ep})))
            ;
        }

        // @see Eqs. (24), (25) in \ref Hoang:2006pd
        double Lam(double u){
          const complex l = complex{
            log((1 + sqrt(1 - 4/u))/(1 - sqrt(1 - 4/u))),
            - pi
          };
          return
            -5./2. + 2./u - (1 + 2/u)*sqrt(1 - 4/u)*real(l)
            - (1 + 1/(2*u))*4/u*real(l*l)
            ;
        }

        //! FA contribution
        /**
         * @see Eq. (B.1) in \ref Grzadkowski:1986pm, Eq. (22) in \ref Hoang:2006pd
         * vector/axialvector indices are considered instead of left/right
         * The overall factor alpha/(4*pi) has been
         * drawn out of the Wilson coefficient
         */
        template<coupling C>
        double FA(parameters const & p);
        template<>
        double FA<vector>(parameters const & p){
          return
            + e*(v_e*v_e + a_e*a_e)*rho(p.s/(mZ*mZ))
            - sqrt(cw2/sw2)*a_e*Lam(p.s/(mW*mW))
            ;
        }

        template<>
        double FA<axial>(parameters const & p){
          return
            + e*2*v_e*a_e*rho(p.s/(mZ*mZ))
            - sqrt(cw2/sw2)*a_e*Lam(p.s/(mW*mW))
            ;
        }

        //! FZ contribution
        /**
         * @see Eq. (B.1) in \ref Grzadkowski:1986pm, Eq. (23) in \ref Hoang:2006pd
         * vector/axialvector indices are considered instead of left/right
         * The overall factor alpha/(4*pi) has been
         * drawn out of the Wilson coefficient
         */
        template<coupling C>
        double FZ(parameters const & p);
        template<>
        double FZ<vector>(parameters const & p){
          return
            + v_e*(v_e*v_e + 3*a_e*a_e)*rho(p.s/(mZ*mZ))
            + (v_e + (1 - 4*cw2)*a_e)/(4*sw2)*rho(p.s/(mW*mW))
            - cw2/sw2*a_e*Lam(p.s/(mW*mW))
            ;
        }

        template<>
        double FZ<axial>(parameters const & p){
          return
            + a_e*(a_e*a_e + 3*v_e*v_e)*rho(p.s/(mZ*mZ))
            + (v_e + (1 - 4*cw2)*a_e)/(4*sw2)*rho(p.s/(mW*mW))
            - cw2/sw2*a_e*Lam(p.s/(mW*mW))
            ;
        }

      }

      template<coupling c>
      double hEW_ee(parameters const & p){
        double photon_contr = FA<c>(p)*p.e_q;
        if(p.production == production_channel::photon_only) return photon_contr;
        const double v_q = p.c_q[vector];
        return photon_contr + FZ<c>(p)*v_q*p.s*Z_prop(p.s);
      }



      /*********** self energies **********/
      //! Electroweak self-energy corrections
      /**
       * @see Eq. (16) in \ref Hoang:2006pd
       * The pure pure QED part (fermionic contribution to photon selfenergy)
       * is moved to Cv_QED
       */
      template<coupling c>
      double hEW_selfenergy(parameters const & p);

      namespace {

        /**
         * @see Eq. (B.4) in \ref Grzadkowski:1986pm
         * the overall factor alpha/(4*pi) has been
         * drawn out of the Wilson coefficient
         */

        enum spin_type{
          boson, fermion
        };

        template<spin_type T>
        double PiAA_bare(double s, parameters const & p);

        template<spin_type T>
        double PiAA(double s, parameters const & p){
          double const s0 = p.mu_alpha*p.mu_alpha;
          return PiAA_bare<T>(s, p) - s/s0*PiAA_bare<T>(s0, p);
        }

        constexpr int num_families = 3;
        constexpr double sum_eq2_leptons = 3;

        double A_0(double m);
        double B_0(double s, double m1, double m2);

        template<>
        double PiAA_bare<boson>(double s, parameters const &){
          return
            -(3*s + 4*mW*mW)*B_0(s, mW, mW)
            + 4*mW*mW*B_0(0, mW, mW)
            ;
        }

        //! sum over light quark charges squared
        constexpr double sum_eq2(int nl){
          return (nl==1)
            ?e_u*e_u
            :(
                + (nl/2)*e_u*e_u
                + ((nl+1)/2)*e_d*e_d
            );
        }

        template<>
        double PiAA_bare<fermion>(double s, parameters const & p){
          // charge sum over leptons - treat all as massless
          return 4./3.*(
              + (QCD::d_r*sum_eq2(p.nl) + sum_eq2_leptons)*s*(
                  B_0(s,0,0) - 1./3.
              )
              + QCD::d_r*p.e_q*p.e_q*(
                  + (s + 2*p.m*p.m)*B_0(s,p.m,p.m)
                  - 2*p.m*p.m*B_0(0,p.m,p.m)
                  - 1./3.*s
              )
          );
        }

        double PiZZ_b(double s, parameters const & p){
          const double mH = p.m_Higgs;
          return (
              + (
                  + (1 - cw2*(4 + 36*cw2))*s
                  + (20 - cw2*(32 + 48*cw2))*mW*mW
              )*s*B_0(s, mW, mW)
              + (4 + cw2*(-16 + 48*cw2))*s*mW*mW*B_0(0, mW, mW)
              - 4./3.*(cw2 - sw2)*s*s
              + (
                  + pow(s + mZ*mZ - mH*mH, 2)
                  + 8*s*mZ*mZ
              )*B_0(s, mZ, mH)
              + (2*s - mZ*mZ + mH*mH)*mZ*mZ*B_0(0, mZ, mZ)
              + (2*s - mH*mH + mZ*mZ)*mH*mH*B_0(0, mH, mH)
              - pow(mZ*mZ - mH*mH, 2)
          )/(12*sw2*cw2*s);
        }

        //! sum over light quark vector Z couplings squared
        double sum_vq2(int nl){
          return (nl==1)
            ?v_u*v_u
            :(
                + (nl/2)*v_u*v_u
                + ((nl+1)/2)*v_d*v_d
            );
        }

        //! sum over light quark axial Z couplings squared
        double sum_aq2(int nl){
          return (nl==1)
            ?a_u*a_u
            :(
                + (nl/2)*a_u*a_u
                + ((nl+1)/2)*a_d*a_d
            );
        }

        double PiZZ_f(double s, parameters const & p){
          // Z coupling sums over leptons - treat all as massless
          static const double sum_a2_v2_leptons = num_families*(
              v_e*v_e + v_nu*v_nu + a_e*a_e + a_nu*a_nu
          );
          const double v_q = p.c_q[vector];
          const double a_q = p.c_q[axial];
          return
            + 4./3.*(
                + QCD::d_r*(sum_vq2(p.nl) + sum_aq2(p.nl))
                + sum_a2_v2_leptons
            )*s*(B_0(s,0,0) - 1./3.)
            + QCD::d_r*(
                + 4./3.*(v_q*v_q + a_q*a_q)*(
                    + (s + 2*p.m*p.m)*B_0(s,p.m,p.m)
                    - 2*p.m*p.m*B_0(0,p.m,p.m)
                    - 1./3.*s
                )
                - 1./(2.*sw2*cw2)*p.m*p.m*B_0(s,p.m,p.m)
            )
            ;
        }

        double PiZA_b(double s, parameters const &){
          return 1./(3.*sqrt(sw2*cw2))*(
              + (-(1./2. + 9*cw2)*s - 4*mW*mW*(1 + 3*cw2))*B_0(s, mW, mW)
              + 2*(6*cw2-1)*mW*mW*B_0(0, mW, mW)
              - 1./3.*s
          );
        }

        //! sum over light quark charges * vector Z couplings
        double sum_eq_vq(int nl){
          return (nl==1)
            ?e_u*v_u
            :(
                + e_u*(nl/2)*v_u
                + e_d*((nl+1)/2)*v_d
            );
        }

        double PiZA_f(double s, parameters const & p){
          const double v_q = p.c_q[vector];
          return 4./3.*(
              + (QCD::d_r*sum_eq_vq(p.nl) + num_families*e*v_e)*s*(
                  B_0(s,0,0) - 1./3.
              )
              + QCD::d_r*p.e_q*v_q*(
                  + (s + 2*p.m*p.m)*B_0(s,p.m,p.m)
                  - 2*p.m*p.m*B_0(0,p.m,p.m)
                  - 1./3.*s
              )
          );
        }

        double PiWW_b(double s, parameters const & p){
          const double mH = p.m_Higgs;
          return (
              + (
                  + (1 - 40*cw2)*s*s
                  + (-16*cw2 - 54 + 10/cw2)*mW*mW*s
                  + sw2*sw2*(1 + 8*cw2)*pow(mZ, 4)
              )*B_0(s, mW, mZ)
              + (pow(s + mW*mW - mH*mH, 2) + 8*s*mW*mW)*B_0(s, mW, mH)
              + 8*sw2*(-5*s*s - 2*s*mW*mW + pow(mW, 4))*B_0(s, 0, mW)
              + (20*s - 2*mW*mW + mZ*mZ + mH*mH)*mW*mW*B_0(0, mW, mW)
              + (8*cw2+1)*(2*s-sw2*mZ*mZ)*mZ*mZ*B_0(0, mZ, mZ)
              + (mW*mW - mH*mH + 2*s)*mH*mH*B_0(0, mH, mH)
              - 4./3.*s*s - sw2*(sw2 + 8*cw2)*pow(mZ, 4) - pow(mW*mW - mH*mH, 2)
          )/(12*sw2*s);
        }

        bool odd(int n){
          return n % 2;
        }

        double PiWW_f(double s, parameters const & p){
          // light doublets
          double result = (num_families + QCD::d_r*(p.nl/2))*s*(
              B_0(s, 0, 0) - 1./3.
          );
          // heavy-light doublet
          // only contributes if the complete doublet (both flavours) exists
          // in the theory with nl + 1 active flavours
          if(odd(p.nl)){
            result += QCD::d_r*(
                + (s - p.m*p.m/2*(1 + p.m*p.m/s))*B_0(s, 0, p.m)
                + (-1 + p.m*p.m/(2*s))*p.m*p.m*B_0(0, p.m, p.m)
                + pow(p.m, 4)/(2*s) - s/3
            );
          }
          return result/(3*sw2);
        }

        double PiZZ_bare(double s, parameters const & p){
          if(p.m < mZ) return PiZZ_f(s, p);
          return PiZZ_b(s, p) + PiZZ_f(s, p);
        }
        double PiZA_bare(double s, parameters const & p){
          if(p.m < mZ) return PiZA_f(s, p);
          return PiZA_b(s, p) + PiZA_f(s, p);
        }
        double PiWW_bare(double s, parameters const & p){
          if(p.m < mZ) return PiWW_f(s, p);
          return PiWW_b(s, p) + PiWW_f(s, p);
        }

        /*********** renormalised self energies **********/
        /**
         * @see Eq. (B.5) in \ref Grzadkowski:1986pm
         * Note that we use the running QED coupling
         * renormalised at mu_alpha. Hence, Pi^{(AA)}'(0)
         * is replaced by Pi^{(AA)}(mu_alpha^2)/mu_alpha^2
         */

        double delta_(parameters const & p){
          return 1/(4*sw2)*(
              PiWW_bare(mW*mW, p)/(mW*mW) - PiZZ_bare(mZ*mZ, p)/(mZ*mZ)
          );
        }

        double delta_Z_ZA(parameters const & p){
          return - 4*sqrt(sw2*cw2)*delta_(p) + 2/(mZ*mZ)*PiZA_bare(0, p);
        }

        double delta_Z_Z(parameters const & p){
          double const s0 = p.mu_alpha*p.mu_alpha;
          return
            - (PiAA_bare<fermion>(s0, p) + PiAA_bare<boson>(s0, p))/s0
            - (sw2 - cw2)/sqrt(sw2*cw2)*delta_Z_ZA(p);
        }

        double PiZZ(double s, parameters const & p){
          return
            PiZZ_bare(s, p) - PiZZ_bare(mZ*mZ, p)
            + (s - mZ*mZ)*delta_Z_Z(p);
        }

        double PiZA(double s, parameters const & p){
          return PiZA_bare(s, p) - PiZA_bare(0, p) + s*delta_Z_ZA(p);
        }

      }

      template<coupling c>
      double hEW_selfenergy(parameters const & p){
        const double v_q = p.c_q[vector];
        static constexpr int only_vector = c == vector;
        double photon_contr = - only_vector*e*p.e_q*PiAA<boson>(p.s, p)/p.s;
        if(p.production == production_channel::photon_only){
          return photon_contr;
        }
        return photon_contr
          - c_e<c>()*v_q*PiZZ(p.s, p)*p.s*Z_prop(p.s)*Z_prop(p.s)
          - (only_vector*e*v_q + c_e<c>()*p.e_q)*PiZA(p.s, p)*Z_prop(p.s)
          ;
      }

      //! absorptive correction
      /*
       * cf. eq. (14) in \ref Hoang:2004tg
       * note that our normalisation differs by a factor
       * of (m/(2*\pi*\alpha))^2
       *
       * actual formulas are by Thomas Rauh, so we can distinguish
       * between the contribution from self-energy insertions ("Zt")
       * and the rest ("bare"). See also the NNLO non-resonant paper,
       * eqs. (48), (49), (75), (76)
       */
      template<int N, C_abs_tag tag>
      double Cv_abs_photon(double x, parameters const & p);
      template<int N, C_abs_tag tag>
      double Cv_abs_Z(double x, parameters const & p);
      template<int N, C_abs_tag tag>
      double Cv_abs_box(double x, parameters const & p);
      template<int N, C_abs_tag tag>
      double Ca_abs_Z(double x, parameters const & p);
      template<int N, C_abs_tag tag>
      double Ca_abs_box(double x, parameters const & p);

      template<>
      double Ca_abs_Z<2,Zt>(double x, parameters const & p){
        const double v_q = p.c_q[vector];
        const double a_q = p.c_q[axial];
        return pi/(8.*sw2)*a_e*(1.-x)*p.s*Z_prop(p.s)*(
            3.*v_q*(1. + x*(1. + 2.*x)) + a_q*(1. + x*(1. - 2.*x))
        )/x;
      }

      template<>
      double Ca_abs_Z<2,Zt_ep>(double x, parameters const & p){
        const double v_q = p.c_q[vector];
        const double a_q = p.c_q[axial];
        return pi/(4.*sw2)*a_e*(1.-x)*p.s*Z_prop(p.s)*(
            + a_q*(1. - x)*(1. + x) + v_q * (2. - x*(1. - x))
            + (
                3.*v_q*(1. + x*(1. + 2.*x)) + a_q*(1. + x*(1. - 2.*x))
            )*log(p.mu_w/(p.m*(1.-x)))
        )/x;
      }

      template<>
      double Ca_abs_Z<2,bare>(double x, parameters const & p){
        static const double sw = sqrt(sw2);
        static const double cw = sqrt(cw2);
        return pi/(24.*sw2)*a_e*(1.-x)*(1.-x)*p.s*Z_prop(p.s)*(
            + (a_d + v_d)*sw*x*(-1. + x*(5. + 2.*x))
            + cw*(-2. + x*(-7. + x*(-3. + 2.*x)))
        )/(x*x*(1. + x)*sw);
      }

      template<>
      double Ca_abs_box<2, Zt>(double, parameters const &){
        return 0.;
      }

      template<>
      double Ca_abs_box<2, Zt_ep>(double, parameters const &){
        return 0.;
      }

      template<>
      double Ca_abs_box<2, bare>(double x, parameters const & p){
        const double cancel_s_dep = p.s/(4*p.m*p.m);
        return pi/(48.*sw2*sw2)*cancel_s_dep*(
            23. - 1./(x*x) + 2./x - 24.*atanh(1. - x)/(1. - x)
        );
      }

      template<>
      double Cv_abs_photon<2,Zt>(double x, parameters const & p){
        return 3.*pi*e/(8.*sw2)*p.e_q*(x-1)*(1 + x*(1 + 2*x))/x;
      }

      template<>
      double Cv_abs_photon<2,Zt_ep>(double x, parameters const & p){
        return e*pi/(4.*sw2)*p.e_q*(x-1)*(
            2. - x*(1-x) + 3.*(1 + x*(1 + 2*x))*log(p.mu_w/(p.m*(1-x)))
        )/x;
      }

      template<>
      double Cv_abs_photon<2,bare>(double x, parameters const &){
        return e*pi/(24.*sw2)*(1-x)*(1-x)*(
            e_d*x*(1. - x*(5. + 2.*x)) + 2.*T3_u*(2. + x*(7. + x*(3.-2.*x)))
        )/(x*x*(1+x));
      }

      template<>
      double Cv_abs_Z<2,Zt>(double x, parameters const & p){
        return -v_e/a_e*Ca_abs_Z<2,Zt>(x, p);
      }

      template<>
      double Cv_abs_Z<2,Zt_ep>(double x, parameters const & p){
        return -v_e/a_e*Ca_abs_Z<2,Zt_ep>(x, p);
      }

      template<>
      double Cv_abs_Z<2,bare>(double x, parameters const & p){
        return -v_e/a_e*Ca_abs_Z<2,bare>(x, p);
      }

      template<>
      double Cv_abs_box<2, Zt>(double, parameters const &){
        return 0.;
      }

      template<>
      double Cv_abs_box<2, Zt_ep>(double, parameters const &){
        return 0.;
      }

      template<>
      double Cv_abs_box<2, bare>(double x, parameters const & p){
        return -Ca_abs_box<2, bare>(x, p);
      }

      template<int N, C_abs_tag tag>
      double Cv_abs_helper(parameters const & p){
        if(p.m < mZ) return 0.;
        const double x = mW*mW/(p.m*p.m);
        double photon_contr = Cv_abs_photon<N, tag>(x, p);
        if(p.production == production_channel::photon_only) return photon_contr;
        return photon_contr + Cv_abs_Z<N, tag>(x, p) + Cv_abs_box<N, tag>(x, p);
      }

      template<int N, C_abs_tag tag>
      double Ca_abs_helper(parameters const & p){
        if(p.m < mZ) return 0.;
        const double x = mW*mW/(p.m*p.m);
        if(p.production == production_channel::photon_only) return 0.;
        return Ca_abs_Z<N, tag>(x, p) + Ca_abs_box<N, tag>(x, p);
      }

    }

    template<int N, C_abs_tag tag>
    double Cv_abs(parameters const & p){
      if(p.contributions.Cv_WZ[0] == 0.) return 0.;
      return p.contributions.Cv_WZ[0]*Cv_abs_helper<N, tag>(p);
    }

    template<int N, C_abs_tag tag>
    double Ca_abs(parameters const & p){
      if(p.contributions.Ca_WZ[0] == 0.) return 0.;
      return p.contributions.Ca_WZ[0]*Ca_abs_helper<N, tag>(p);
    }

    // enforce instantiation
    template
    double Cv_abs<2, Zt>(parameters const & p);
    template
    double Ca_abs<2, Zt>(parameters const & p);
    template
    double Cv_abs<2, Zt_ep>(parameters const & p);
    template
    double Ca_abs<2, Zt_ep>(parameters const & p);
    template
    double Cv_abs<2, bare>(parameters const & p);
    template
    double Ca_abs<2, bare>(parameters const & p);

    template<>
    double Cv_abs<2>(parameters const & p){
      return Cv_abs<2, Zt>(p) + Cv_abs<2, bare>(p);
    }

    template<>
    double Ca_abs<2>(parameters const & p){
      return Ca_abs<2, Zt>(p) + Ca_abs<2, bare>(p);
    }

    /********** sum of electroweak corrections **********/

    namespace{
      //! Axialvector/vector electroweak corrections
      /**
       * @see Eq. (15) in \ref Hoang:2006pd
       * hEW<vector> = (hEW_R + hEW_L)/2
       * hEW<axial> = (hEW_L - hEW_R)/2
       */
      template<coupling c>
        double hEW(parameters const & p){
        return
          hEW_selfenergy<c>(p) + hEW_ee<c>(p) + hEW_QQbar<c>(p) + hEW_box<c>(p)
          ;
      }

    }

    /****** separation into pure QED corrections and rest ******/
    template<int N>
    complex Cv_QED(parameters const & p);
    template<>
    complex Cv_QED<2>(parameters const & p){
      if(p.contributions.Cv_QED[0] == 0.) return 0.;
      return p.contributions.Cv_QED[0]*p.e_q*(
          p.e_q*Cv_p(p)*cv_QED_QQbar<2>(p) - e*PiAA<fermion>(p.s, p)/p.s
      );
    }

    template<int N>
    complex Ca_QED(parameters const & p);
    template<>
    complex Ca_QED<2>(parameters const & p){
      if(p.contributions.Ca_QED[0] == 0.) return 0.;
      return p.contributions.Ca_QED[0]*p.e_q*p.e_q*Ca_p(p)*cv_QED_QQbar<2>(p);
    }

    template<int N>
    complex Cv_WZ(parameters const & p);
    template<>
    complex Cv_WZ<2>(parameters const & p){
      if(p.contributions.Cv_WZ[0] == 0. || p.m < mZ/2) return 0.;
      return complex{p.contributions.Cv_WZ[0]*hEW<vector>(p), Cv_abs<2>(p)};
    }

    template<int N>
    complex Ca_WZ(parameters const & p);
    template<>
    complex Ca_WZ<2>(parameters const & p){
      if(p.contributions.Ca_WZ[0] == 0. || p.m < mZ/2) return 0.;
      return complex{-p.contributions.Ca_WZ[0]*hEW<axial>(p), Ca_abs<2>(p)};
    }


    //! Leading electroweak correction to hard matching coefficient
    /**
     * @see \ref Grzadkowski:1986pm, \ref Hoang:2004tg, \ref Hoang:2006pd
     * cf. eqs. (12)--(14) in \ref Hoang:2006pd, eq. (14) in \ref Hoang:2004tg
     * Note that our normalisation of Ca, Cv differs by a factor
     * of m^2/(\alpha*\pi)*4*\pi/\alpha = 4*m^2/\alpha^2, and in our convention
     * C^{ew}_{V} = (h^{ew}_{L} + h^{ew}_{R})/2 = h^{ew}_{V}
     * C^{ew}_{A} = (h^{ew}_{R} - h^{ew}_{L})/2 = -h^{ew}_{A}
     * when neglecting the pure QED corrections
     * (note the additional minus sign in the axial part)
     * We also divide C^{ew}_{V/A} into pure QED corrections
     * and corrections with at least one W, Z, or Goldstone boson
     *
     * To be consistent with our leading-order prescription
     * we set s -> 4*m^2 in the box and q qbar vertex corrections,
     * but keep the full dependence in the e^+ e^-  and self-energy
     * corrections.
     *
     * TODO: NNNLO correction
     */
    template<int N>
    complex Cv_EW(parameters const & p){
      return Cv_QED<N>(p) + Cv_WZ<N>(p);
    }
    template<int N>
    complex Ca_EW(parameters const & p){
      return Ca_QED<N>(p) + Ca_WZ<N>(p);
    }
    template
    complex Cv_EW<2>(parameters const & p);
    template
    complex Ca_EW<2>(parameters const & p);

  }
}
