/**
 * @file
 * @brief Implementation of potential insertion functions
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#include "insertions.hpp"

#include <cmath>

#include "complex.hpp"
#include "parameters.hpp"
#include "QCD/colour.hpp"
#include "zeta.hpp"
#include "harmonic_sums.hpp"
#include "psi.hpp"
#include "log_lam.hpp"
#include "hypergeometric.hpp"

namespace QQbar_threshold{
  namespace detail{
    using QCD::c_F;
    using namespace zeta;
    using psi::psi;
    using harmonic_sums::harmonic_sum;
    using namespace complex;
    using hypergeometric::PFQ;

    /*************************************************************************/
    /*                       Coulomb single insertions                       */
    /*************************************************************************/

    template<>
    complex insertion<1>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(p.m/(4*pi),2)*(
	  + 1./2. - log_lam(lam, p)
	  - harmonic_sum<1>(-1-lam) + lam*psi<1>(-lam)
      );
    }

    template<>
    complex insertion_1<1>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(p.m/(4*pi),2)*(
	  + 1./3. - log_lam(lam, p)
	  - harmonic_sum<1>(-1-lam) + lam*psi<1>(-lam)
      );
    }

    template<>
    complex insertion<1+ep>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(p.m/pi,2)*(
	  - 1./48.
	  + 5./192.*zeta_2
	  + PFQ<4>({1,1,1,1},{2,2,1-lam},1)/4
	  + lam*psi<2>(-lam)/16
	  + log_lam(lam,p)*(
	      + log_lam(lam,p)/16
	      - lam*psi<1>(-lam)/8
	      + harmonic_sum<1>(-1-lam)/8
	  )
	  + pow(harmonic_sum<1>(-1-lam),2)/16
	  + psi<1>(-lam)*(-1./16. - lam*harmonic_sum<1>(-1-lam)/8)
      );
    }


    template<>
    complex insertion<1+2*ep>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(p.m/pi,2)*(
	  + 49./16.
	  + zeta_2/64
	  + 13./16.*zeta_3
	  + gamma_Euler*(
	      - 1
	      + zeta_2
	  )
	  + (
	      - 1
	      + gamma_Euler/2
	      + harmonic_sum<1>(-1-lam)/2
	  )/lam
	  + PFQ<4>({1,1,1,1},{2,2,1-lam},1) * (-1 + gamma_Euler/2 + 1/(2*lam))
	  -  2*PFQ<5>({1,1,1,1,1},{2,2,2,1-lam},1)
	  + 5./12.*psi<2>(-lam)
	  + insertion_sum1<1+2*ep>(lam)/4
	  + harmonic_sum<1>(-1-lam)*(
	      - 1
	      - 9./8.*zeta_2
	      + (
		  - harmonic_sum<1>(-1-lam)/4
		  - log_lam(lam,p)/4
	      )*log_lam(lam,p)
	      - PFQ<4>({1,1,1,1},{2,2,1-lam},1)/2
	      - 3./4.*psi<1>(-lam)
	      - pow(harmonic_sum<1>(-1-lam),2)/12
	  )
	  + lam*(
	      - gamma_Euler*zeta_3
	      + zeta_3*harmonic_sum<1>(-1-lam)
	      + (
		  + harmonic_sum<1>(-1-lam)/4
		  - log_lam(lam,p)/4
	      )*psi<2>(-lam)
	      - psi<3>(-lam)/6
	      + psi<1>(-lam)*(
		  + 2-gamma_Euler/2
		  + 9./8.*zeta_2
		  + harmonic_sum<1>(-1-lam)*(
		      - 1./2.
		      + harmonic_sum<1>(-1-lam)/4
		      + log_lam(lam,p)/2
		  )
		  + psi<1>(-lam)/4
		  + pow(log_lam(lam,p),2)/4
	      )
	  )
	  + log_lam(lam,p)*(
	      - zeta_2/8
	      - PFQ<4>({1,1,1,1},{2,2,1-lam},1)
	      + psi<1>(-lam)/4
	      - pow(log_lam(lam,p),2)/12
	  )
      );
    }


    template<>
    complex insertion<1+3*ep>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(p.m/pi,2)/lam*(
	  - 6
	  + 3*gamma_Euler
	  + 3*harmonic_sum<1>(-1-lam)
	  + log_lam(lam,p)*(
	      + 6
	      - 3*gamma_Euler
	      - 3*harmonic_sum<1>(-1-lam)
	      - 3*PFQ<4>({1,1,1,1},{2,2,1-lam},1)
	  )
	  + 3*PFQ<4>({1,1,1,1},{2,2,1-lam},1)
	  + lam*(
	      + 285./16.
	      + 93./64.*zeta_2
	      + 81./16.*zeta_3
	      + 3./2.*insertion_sum1<1+2*ep>(lam)
	      + insertion_sum1<1+3*ep>(lam)/16
	      + (
		  - 12
		  + 9./2.*zeta_2
	      )*PFQ<4>({1,1,1,1},{2,2,1-lam},1)
	      + gamma_Euler*(
		  - 6
		  + 6*zeta_2
		  + 3*PFQ<4>({1,1,1,1},{2,2,1-lam},1)
	      )
	      - 12*PFQ<5>({1,1,1,1,1},{2,2,2,1-lam},1)
	      + harmonic_sum<1>(-1-lam)*(
		  - 9
		  - 9./2.*zeta_2
		  + (
		      - 3./2.
		      + 9./8.*zeta_2
		      - harmonic_sum<1>(-1-lam)/2
		  )*harmonic_sum<1>(-1-lam)
		  - 3*PFQ<4>({1,1,1,1},{2,2,1-lam},1)
		  - 9./2.*psi<1>(-lam)
	      )
	      + (3./2. - 9./8.*zeta_2)*psi<1>(-lam)
	      + 5./2.*psi<2>(-lam)
	      + lam*(
		  - 6*gamma_Euler*zeta_3
		  + 6*zeta_3*harmonic_sum<1>(-1-lam)
		  + psi<1>(-lam)*(
		      + 15
		      - 3*gamma_Euler
		      + 9./2.*zeta_2
		      + harmonic_sum<1>(-1-lam)*(
			  - 9./4.*zeta_2
			  + 3./2.*harmonic_sum<1>(-1-lam)
		      )
		      + 3./2.*psi<1>(-lam)
		  )
		  + (
		      - 3./2.
		      + 9./8.*zeta_2
		      + 3./2.*harmonic_sum<1>(-1-lam)
		  )*psi<2>(-lam)
		  - psi<3>(-lam)
	      )
	      + 221./30720.*pow(pi,4)
	      + log_lam(lam,p)*(
		  - 18
		  - 5*zeta_3
		  - 3./2.*insertion_sum1<1+2*ep>(lam)
		  + gamma_Euler*(
		      + 6
		      - 6*zeta_2
		      - 3*PFQ<4>({1,1,1,1},{2,2,1-lam},1)
		  )
		  + 6*PFQ<4>({1,1,1,1},{2,2,1-lam},1)
		  + 12*PFQ<5>({1,1,1,1,1},{2,2,2,1-lam},1)
		  - 5./2.*psi<2>(-lam)
		  + harmonic_sum<1>(-1-lam)*(
		      6
		      + 27./4.*zeta_2
		      + (
			  3./4.*harmonic_sum<1>(-1-lam)
			  + log_lam(lam,p)/2
		      )*log_lam(lam,p)
		      + 3*PFQ<4>({1,1,1,1},{2,2,1-lam},1)
		      + 9./2.*psi<1>(-lam)
		      + pow(harmonic_sum<1>(-1-lam),2)/2
		  )
		  + lam*(
		      + 6*gamma_Euler*zeta_3
		      - 6*zeta_3*harmonic_sum<1>(-1-lam)
		      + (
			  - 3./2.*harmonic_sum<1>(-1-lam)
			  + 3./4.*log_lam(lam,p)
		      )*psi<2>(-lam)
		      + psi<3>(-lam)
		      + psi<1>(-lam)*(
			  - 12
			  + 3*gamma_Euler
			  - 27./4.*zeta_2
			  + harmonic_sum<1>(-1-lam)*(
			      + 3
			      - 3./2.*harmonic_sum<1>(-1-lam)
			      - 3./2.*log_lam(lam,p)
			  )
			  - 3./2.*psi<1>(-lam)
			  - pow(log_lam(lam,p),2)/2
		      )
		  )
		  + log_lam(lam,p)*(
		      + 3./8.*zeta_2
		      + 3*PFQ<4>({1,1,1,1},{2,2,1-lam},1)
		      - 3./4.*psi<1>(-lam)
		      + pow(log_lam(lam,p),2)/8
		  )
	      )
	  )
      );
    }


    /***************************************************/
    /* Coulomb Double Insertions                       */
    /***************************************************/


    template<>
    complex insertion<1, 1>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(p.m/pi,2)/(c_F*pi*p.alpha_s)*lam*(
	  psi<1>(1-lam)/64 - lam*psi<2>(1-lam)/128
      );
    }


    template<>
    complex insertion<1, 1+ep>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(p.m/pi,2)/(c_F*pi*p.alpha_s)*lam*(
	  - insertion_sum1<1, 1+ep>(lam)/32
	  + log_lam(lam,p)*(
	      - psi<1>(1-lam)/32 + lam*psi<2>(1-lam)/64
	  )
      );
    }

    template<>
    complex insertion<1+ep, 1+ep>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(p.m/pi,2)/(8*c_F*pi*p.alpha_s)*lam*(
	  + insertion_sum1<1+ep, 1+ep>( lam)/2
	  + log_lam(lam, p)*(
	      insertion_sum1<1, 1+ep>( lam)
	      + log_lam(lam, p)*(
		  + psi<1>(-lam)/2
		  - (lam*psi<2>(-lam))/4
	      )
	  )
      );
    }


    template<>
    complex insertion<1, 1+2*ep>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(p.m/pi,2)/(c_F*pi*p.alpha_s)*(
	  + insertion_sum1<1, 1+2*ep>(lam)/8
	  + lam*(
	      + insertion_sum2<1, 1+2*ep>(lam)/8
	      + log_lam(lam,p)*(
		  + insertion_sum1<1, 1+ep>(lam)/8
		  + log_lam(lam,p)*(
		      + psi<1>(1-lam)/16
		      - lam*psi<2>(1-lam)/32
		  )
	      )
	  )
      );
    }


    template<>
    complex insertion<1+ep, 1+2*ep>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(p.m/pi,2)/(c_F*pi*p.alpha_s)*(
	  - log_lam(lam,p)*insertion_sum1<1, 1+2*ep>(lam)/4
	  + insertion_sum1<1+ep, 1+2*ep>(lam)/4
	  + lam*(
	      - insertion_sum3<1+ep, 1+2*ep>(lam)/32
	      + log_lam(lam,p)*(
		  - insertion_sum2<1+ep, 1+2*ep>(lam)/4
		  + log_lam(lam,p)*(
		      - 3./8.*insertion_sum1<1, 1+ep>(lam)
		      + log_lam(lam,p)*(
			  - psi<1>(1-lam)/8
			  + lam*psi<2>(1-lam)/16
		      )
		  )
	      )
	  )
      );
    }


    /***************************************************/
    /* Coulomb Triple Insertions                       */
    /***************************************************/


    template<>
    complex insertion<1, 1, 1>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(lam*p.m/(pi*pi*c_F*p.alpha_s),2)*(
	  - psi<2>(1-lam)/512 + lam*psi<3>(1-lam)/1536
      );
    }


    template<>
    complex insertion<1, 1, 1+ep>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(lam*p.m/(pi*pi*c_F*p.alpha_s),2)*(
	  - insertion_sum1<1, 1, 1+ep>(lam)/128
	  + log_lam(lam, p)*(
	      psi<2>(1-lam)/256 - lam*psi<3>(1-lam)/768
	  )
      );
    }


    template<>
    complex insertion<1+ep, 1, 1+ep>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(lam*p.m/(pi*pi*c_F*p.alpha_s),2)*(
	  + insertion_sum1<1+ep, 1, 1+ep>(lam)/64
	  + log_lam(lam, p)*(
	      + insertion_sum1<1, 1, 1+ep>(lam)/32
	      + log_lam(lam, p)*(- psi<2>(1-lam)/128 + lam*psi<3>(1-lam)/384)
	  )
      );
    }


    template<>
    complex insertion<1, 1+ep, 1>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(lam*p.m/(pi*pi*c_F*p.alpha_s),2)*(
	  + insertion_sum1<1, 1+ep, 1>(lam)/128
	  - psi<3>(1-lam)/768
	  + lam*(
	      + psi<1>(1-lam)*psi<2>(1-lam)/256
	      + psi<4>(1-lam)/1536
	      + psi<3>(1-lam)*(
		  - log_lam(lam, p)/768
		  + psi<0>(1-lam)/384
		  - harmonic_sum<1>(-lam)/768
	      )
	  )
	  + psi<2>(1-lam)*(
	      + log_lam(lam, p)/256
	      - psi<0>(1-lam)/128
	      + harmonic_sum<1>(-lam)/256
	  )
      );
    }


    template<>
    complex insertion<1, 1+ep, 1+ep>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(lam*p.m/(pi*pi*c_F*p.alpha_s),2)*(
	  + insertion_sum1<1, 1+ep, 1+ep>(lam)/(128*pow(lam,2))
	  + log_lam(lam, p)*(
	      + insertion_sum2<1, 1+ep, 1+ep>(lam)/64
	      + log_lam(lam, p)*(
		  - psi<2>(1-lam)/128
		  + lam*psi<3>(1-lam)/384
	      )
	  )
      );
    }


    template<>
    complex insertion<1+ep, 1+ep, 1+ep>(
		      complex lam,
		      parameters const & p
		      ){
      return pow(lam*p.m/(pi*pi*c_F*p.alpha_s),2)*(
	  + insertion_sum1<1+ep, 1+ep, 1+ep>(lam)/32
	  + log_lam(lam, p)*(
	      + insertion_sum2<1+ep, 1+ep, 1+ep>(lam)/32
	      + log_lam(lam, p)*(
		  + insertion_sum3<1+ep, 1+ep, 1+ep>(lam)/32
		  + log_lam(lam, p)*(
		      + psi<2>(1-lam)/64
		      - lam*psi<3>(1-lam)/192
		  )
	      )
	  )
      );
    }



    /***************************************************/
    /* Non-Coulomb Single Insertions                   */
    /***************************************************/


    template<>
    complex insertion<0>(complex lam, parameters const & p){
      return pow(c_F*p.alpha_s/(4*pi*lam),2)*pow(p.m,4)*(
	  + 1./4.
	  + lam*(
	      - 1./2.
	      + log_lam(lam, p)
	      + harmonic_sum<1>(-lam)
	      + lam*(
		  + 1./4.
		  + (-1 + harmonic_sum<1>(-lam))*harmonic_sum<1>(-lam)
		  + log_lam(lam, p)*(
		      -1
		      + log_lam(lam, p)
		      + 2*harmonic_sum<1>(-lam)
		  )
	      )
	  )
      );
    }

    namespace{
      /** \internal
       * For the RJ insertions we use tag dispatch, so that
       * we don't have to define separate functions for insertions
       * 1/2+1*ep, 1/2+2*ep etc. but instead use a single function
       * for all cases 1/2+a*ep
       */
      template<int IntPow, int EpPow> struct plus {};
      template<int EpPow>
	complex insertion_R0(complex lam, parameters const & p, plus<one_half, EpPow>);
      template<int EpPow>
	complex insertion_R1(complex lam, parameters const & p, plus<one_half, EpPow>);
      template<int EpPow>
	complex insertion_R2(complex lam, parameters const & p, plus<one_half, EpPow>);
      template<int EpPow>
	complex insertion_R0(complex lam, parameters const & p, plus<0, EpPow>);
      template<int EpPow>
	complex insertion_R1(complex lam, parameters const & p, plus<0, EpPow>);
      template<int EpPow>
	complex insertion_R2(complex lam, parameters const & p, plus<0, EpPow>);
    }

    template<ins_type I>
    complex insertion_R0(complex lam, parameters const & p){
      static_assert(I == I%ep + I/ep*ep, "decomposition of I");
      return insertion_R0<I/ep*ep>(lam, p, plus<I%ep, I/ep*ep>{});
    }

    template<ins_type I>
    complex insertion_R1(complex lam, parameters const & p){
      static_assert(I == I%ep + I/ep*ep, "decomposition of I");
      return insertion_R1<I/ep*ep>(lam, p, plus<I%ep, I/ep*ep>{});
    }

    template<ins_type I>
    complex insertion_R2(complex lam, parameters const & p){
      static_assert(I == I%ep + I/ep*ep, "decomposition of I");
      return insertion_R2<I/ep*ep>(lam, p, plus<I%ep, I/ep*ep>{});
    }
    namespace{
      template<int EpPow>
	complex insertion_R0(
	    complex lam,
	    parameters const & p,
	    plus<one_half, EpPow>
	){
	static_assert(EpPow%ep == 0, "EpPow must be an integer multiple of ep");
	static constexpr int a = EpPow/ep;
	return a*c_F*p.alpha_s/(lam*(1+a))*pow(p.m/pi,3)*(
	    - 13./32.*zeta_2-insertion_sum1<one_half+ep>(lam)/4-log(2)/4
	    + pow(log(2),2)/8
	    + psi<1>(1-lam)/4
	    + log_lam(lam, p)*(
		+ 1./4. - log_lam(lam, p)/8 - harmonic_sum<1>(-lam)/4
	    )
	    + (1./2. - psi<0>(1-lam)/4)*harmonic_sum<1>(-lam)
	    + a*(
		- 1./4.
		- 5./16.*zeta_2
		- insertion_sum1<one_half+ep>(lam)/4
		+ psi<1>(1-lam)/4
		+ log_lam(lam, p)*(
		    1./4.-log_lam(lam, p)/8-harmonic_sum<1>(-lam)/4
		)
		+ (1./2. - psi<0>(1-lam)/4)*harmonic_sum<1>(-lam)
	    )
	    + lam*(
		- pow(log(2),3)/4
		- psi<1>(1-lam)/4
		- psi<2>(1-lam)/12
		+ log_lam(lam, p)*(
		    - zeta_2/16
		    - log(2)/2
		    + pow(log(2),2)/4
		    + log_lam(lam, p)*(
			+ 1./4.
			-log_lam(lam, p)/12
			-harmonic_sum<1>(-lam)/4
		    )
		    + (1./2.-harmonic_sum<1>(-lam)/4)*harmonic_sum<1>(-lam)
		)
		+ psi<0>(1-lam)*(
		    - 9./16.*zeta_2
		    - log(2)/2
		    + pow(log(2),2)/4
		    + psi<1>(1-lam)/4
		    + psi<0>(1-lam)*(
			- 1./4. - psi<0>(1-lam)/12 + harmonic_sum<1>(-lam)/4
		    )
		    + (1./2.-harmonic_sum<1>(-lam)/4)*harmonic_sum<1>(-lam)
		)
		+ a*(
		    - psi<1>(1-lam)/4
		    - psi<2>(1-lam)/12
		    + log_lam(lam, p)*(
			- 1./2.
			+ zeta_2/8
			+ log_lam(lam, p)*(
			    + 1./4.
			    - log_lam(lam, p)/12
			    - harmonic_sum<1>(-lam)/4
			)
			+ (1./2. - harmonic_sum<1>(-lam)/4)*harmonic_sum<1>(-lam)
		    )
		    + psi<0>(1-lam)*(
			- 1./2.
			- 3./8.*zeta_2
			+ psi<1>(1-lam)/4
			+ psi<0>(1-lam)*(
			    - 1./4.
			    - psi<0>(1-lam)/12
			    + harmonic_sum<1>(-lam)/4
			)
			+ (1./2.-harmonic_sum<1>(-lam)/4)*harmonic_sum<1>(-lam)
		    )
		)
	    )
	);
      }


      template<int EpPow>
	complex insertion_R1(
	    complex lam,
	    parameters const & p,
	    plus<one_half, EpPow>
	){
	static_assert(EpPow%ep == 0, "EpPow must be an integer multiple of ep");
	static constexpr int a = EpPow/ep;
	return c_F*p.alpha_s/(lam*(1+a))*pow(p.m/(2*pi),3)*(
	    - 2 - a
	    + log(2)
	    + (log_lam(lam, p) + harmonic_sum<1>(-lam))*(1 + a)
	    + lam*(
		+ (pow(log_lam(lam, p),2) - pow(psi<0>(1-lam),2))*(1 + a)
		+ (log_lam(lam, p) + psi<0>(1-lam))*(
		    - 4 - 2*a
		    + log(4)
		    + 2*harmonic_sum<1>(-lam) * (1 + a)
		)
	    )
	);
      }


      template<int EpPow>
	complex insertion_R2(
	    complex lam,
	    parameters const & p,
	    plus<one_half, EpPow>
	){
	static_assert(EpPow%ep == 0, "EpPow must be an integer multiple of ep");
	static constexpr int a = EpPow/ep;
	return c_F*p.alpha_s/(lam*(1+a))*pow(p.m/(2*pi),3)*(
	    -1./2. + lam*(- log_lam(lam, p) - psi<0>(1-lam))
	);
      }

      template<>
	complex insertion_R0<ep>(
	    complex lam,
	    parameters const & p,
	    plus<0, ep>
	){
	double const lmu = 2*log(p.m*p.alpha_s/p.mu);
	return pow(p.m,4)*pow(c_F/pi*p.alpha_s/lam,2)*(
	    - log_lam(lam, p)/16
	    + lam*(
		lam*(
		    + 9./32.*lmu*insertion_sum1<ep>(lam)
		    + insertion_sum2<ep>(lam)/5760
		)
		+ log_lam(lam, p)*(
		    - 1./16.
		    - log_lam(lam, p)/16
		    + lam*(
			+ 7./12.
			- 5./96.*zeta_2
			- log(2)/8
			+ log_lam(lam, p)*(
			    - 3./8.
			    - log_lam(lam, p)/24
			    - harmonic_sum<1>(-lam)/8
			)
			+ (
			    - 3./8.-harmonic_sum<1>(-lam)/8
			)*harmonic_sum<1>(-lam)
		    )
		)
	    )
	);
      }


      template<>
	complex insertion_R1<ep>(
	    complex lam,
	    parameters const & p,
	    plus<0, ep>
	){
	return pow(p.m,4)*pow(c_F/pi*p.alpha_s/lam,2)*(
	    1./48.
	    + lam*(
		- 1./96.
		+ log_lam(lam, p)/16
		+ harmonic_sum<1>(-lam)/16
		+ lam*(
		    - 1./120.
		    + 13./192.*zeta_2
		    + log_lam(lam, p)*(
			- 1./48.
			+ log_lam(lam, p)/16
			+ harmonic_sum<1>(-lam)/8
		    )
		    + (
			- 1./48.
			+ harmonic_sum<1>(-lam)/16
		    )*harmonic_sum<1>(-lam)
		)
	    )
	);
      }



      template<>
	complex insertion_R2<ep>(
	    complex lam,
	    parameters const & p,
	    plus<0, ep>
	){
	return pow(p.m,4)*pow(c_F/pi*p.alpha_s,2)/(48*lam)*(
	    - 1./2. + lam*(1./20. - log_lam(lam, p) - harmonic_sum<1>(-lam))
	);
      }
    }

    //force template instantiations
    template
    complex insertion_R0<ep>(complex lam, parameters const & p);
    template
    complex insertion_R1<ep>(complex lam, parameters const & p);
    template
    complex insertion_R2<ep>(complex lam, parameters const & p);
    template
    complex insertion_R0<one_half+ep>(complex lam, parameters const & p);
    template
    complex insertion_R1<one_half+ep>(complex lam, parameters const & p);
    template
    complex insertion_R2<one_half+ep>(complex lam, parameters const & p);
    template
    complex insertion_R0<one_half+2*ep>(complex lam, parameters const & p);
    template
    complex insertion_R1<one_half+2*ep>(complex lam, parameters const & p);
    template
    complex insertion_R2<one_half+2*ep>(complex lam, parameters const & p);
    // for test
    template
    complex insertion_R0<one_half+7*ep>(complex lam, parameters const & p);
    template
    complex insertion_R1<one_half+7*ep>(complex lam, parameters const & p);
    template
    complex insertion_R2<one_half+7*ep>(complex lam, parameters const & p);



    template<>
    complex insertion<delta>(
	complex lam,
	parameters const & p
    ){
      return p.m*lam/(2*c_F*pi*p.alpha_s)*(
	  1./2. + lam*(1 + lam*psi<1>(1 - lam))
      );
    }



    /* Notation for the Insertion functions of N2*C1 is different from yk   */
    /***************************************************/
    /* Non-Coulomb Double Insertions                   */
    /***************************************************/


    template<>
    complex insertion<1, one_half>(
	complex lam,
	parameters const & p
    ){
      return
	+ 2*(1 - log(2))*insertion<1>(lam,p)
	+ pow(p.m/pi,2)/8*(
	    + zeta_2
	    + (
		- 2
		+ harmonic_sum<1>(-1-lam)
	    )*harmonic_sum<1>(-1-lam)
	    + (
		+ 1
		+ lam*(
		    + 2
		    - 2*harmonic_sum<1>(-1-lam)
		)
	    )*psi<1>(-lam)
	    + log_lam(lam,p)*(
		- 2
		+ 2*harmonic_sum<1>(-1-lam)
		+ log_lam(lam,p)
		- 2*lam*psi<1>(-lam)
	    )
	);
    }


    template<>
    complex insertion<1+ep, one_half>(
	complex lam,
	parameters const & p
    ){
      return
	+ 2*(1 - log(2))*insertion<1+ep>(lam,p)
	+ pow(p.m/pi,2)/8*(
	    - insertion_sum2<1+ep, one_half>(lam)/3
	    + log_lam(lam,p)*(
		- 2*insertion_sum1<1+ep, one_half>(lam)
		+ log_lam(lam,p)*(
		    + 2
		    - 4*harmonic_sum<1>(-1-lam)
		    - 4./3.*log_lam(lam,p)
		    + 4*lam*psi<1>(-lam)
		)
	    )
	);
    }


    template<>
    complex insertion<1, delta>(
	complex lam,
	parameters const & p
    ){
      return p.m*pow(lam/(c_F*pi*p.alpha_s),2)*(
	  1./16. + lam*(psi<1>(1-lam)/8 - lam*psi<2>(1-lam)/16)
      );
    }


    template<>
    complex insertion<1+ep, delta>(
	complex lam,
	parameters const & p
    ){
      return p.m*pow(lam/(c_F*pi*p.alpha_s),2)*(
	  + lam*insertion_sum1<1+ep, delta>(lam)/4
	  + log_lam(lam, p)*(
	      - 1./8.
	      + lam*(-psi<1>(1-lam)/4 + (lam*psi<2>(1-lam))/8)
	  )
      );
    }
  }

}
