/**
 * @file
 * @brief Non-resonant corrections
 *
 * Results and conventions taken from arXiv:1004.2188 (\cite Beneke:2010mp)
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.0.0
 *
 * @section LICENSE
 *
 * Copyright 2015, 2016, 2017 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#include "non-resonant.hpp"

#include <array>
#include <stdexcept>
#include <string>

#include "interpolation_grid.hpp"
#include "constants.hpp"
#include "width.hpp"
#include "zeta.hpp"
#include "Wilson_coefficients.hpp"
#include "log_lam.hpp"

namespace QQbar_threshold{
  namespace detail{
    using zeta::pi;
    namespace{
      const double a_u = axial_Z_coupling(T3_u, e_u);
      const double a_d = axial_Z_coupling(T3_d, e_d);
      const double a_e = axial_Z_coupling(T3_e, e);

      const double v_u = vector_Z_coupling(T3_u, e_u);
      const double v_d = vector_Z_coupling(T3_d, e_d);
      const double v_e = vector_Z_coupling(T3_e, e);

      //! Diagrams contributing to the non-resonant cross section
      /**
       * TODO:
       * Definitions as in fig. 1 of \cite Beneke:2010mp up to a global factor?
       */
      template<int N>
        double h(double x, double y, parameters const & p);

      template<int Start, int End>
        struct sum_h{
        template<typename... Args>
        double operator()(Args... a){
          return h<Start>(a...) + sum_h<Start + 1, End>{}(a...);
        }
      };

      template<int End>
        struct sum_h<End, End>{
        template<typename... Args>
        double operator()(Args&&... a){
          return h<End>(std::forward<Args>(a)...);
        }
      };

      //! First auxiliary function from eq. (14) of \cite Beneke:2010mp
      double CAA(parameters const & p){
        return -pow(e,2)*pow(p.m,2)/(4*p.s);
      }
      //! Second auxiliary function from eq. (14) of \cite Beneke:2010mp
      double CAZ(parameters const & p){
        return e*v_e*pow(p.m,2)/(4*(p.s-pow(mZ,2)));
      }
      //! Third auxiliary function from eq. (14) of \cite Beneke:2010mp
      double CZZ(parameters const & p){
        return -p.s*(pow(v_e, 2)+pow(a_e, 2))*pow(p.m,2)/(4*pow(p.s-pow(mZ,2),2));
      }
      //! Fourth auxiliary function from eq. (14) of \cite Beneke:2010mp
      double CA (parameters const & p){
        return e*pow(p.m,2)/(sw2*p.s);
      }
      //! Fifth auxiliary function from eq. (14) of \cite Beneke:2010mp
      double CZ (parameters const & p){
        return (v_e+a_e)*pow(p.m,2)/(sw2*(p.s-pow(mZ,2)));
      }

      //! h functions in appendix A of \cite Beneke:2010mp
      /**
       *  Instead of performing the integrals directly,
       *  we use interpolation of grids generated with
       *  gen_non-resonant.m
       */
      double integral(integral_type h, double x, double y){
        assert((y <= 1) && (y > x));
        const double yw = (1-y)/(1-x);
        return h_interpolation(h, x, yw);
      }

      template<>
        double h<1>(double x, double y, parameters const & p){
        return (pow(e_u, 2)*CAA(p)-2*v_u*e_u*CAZ(p)+CZZ(p)*pow(v_u, 2))*6*sqrt(2)/pow(pi,2)/sqrt(1-y)
          +(CAA(p)*pow(e_u, 2)+CAZ(p)*(-2*e_u*v_u)+CZZ(p)*pow(v_u, 2))*integral(h1V, x, y)
          +(CAZ(p)*2*e_u*a_u+CZZ(p)*(-2*v_u*a_u))*integral(h1VA, x, y)
          +CZZ(p)*pow(a_u, 2)*integral(h1A, x, y)
          ;
      }

      template<>
        double h<2>(double x, double y, parameters const & p){
        return (CAA(p)*2*e_d*e_u+CAZ(p)*(-2*(e_d*v_u+e_u*(v_d+a_d)))+CZZ(p)*2*(v_d+a_d)*v_u)*integral(h2V, x, y)
          +(CAZ(p)*(-2*a_u*e_d)+CZZ(p)*2*(v_d+a_d)*a_u)*integral(h2A, x, y)
          ;
      }

      template<>
        double h<3>(double x, double y, parameters const & p){
        return (CAA(p)*2*e_u+CAZ(p)*(-2*v_u-2*sqrt(cw2/sw2)*e_u)+CZZ(p)*2*sqrt(cw2/sw2)*v_u)*integral(h3V, x, y)
          +(CAZ(p)*2*a_u+CZZ(p)*(-2*sqrt(cw2/sw2)*a_u))*integral(h3A, x, y)
          ;
      }

      template<>
        double h<4>(double x, double y, parameters const & p){
        return (CA(p)*e_u+CZ(p)*v_u)*integral(h4V, x, y)
          +CZ(p)*a_u*integral(h4A, x, y)
          ;
      }

      template<>
        double h<5>(double x, double y, parameters const &){
        return 1/pow(sw2, 2)*integral(h5, x, y);
      }

      template<>
        double h<6>(double x, double y, parameters const & p){
        return (CA(p)+CZ(p)*sqrt(cw2/sw2))*integral(h6, x, y);
      }

      template<>
        double h<7>(double x, double y, parameters const & p){
        return (CA(p)*e_d+CZ(p)*(v_d+a_d))*integral(h7, x, y);
      }

      template<>
        double h<8>(double x, double y, parameters const & p){
        return (CAA(p)*pow(e_d, 2)+CAZ(p)*(-2*e_d*(v_d+a_d))+CZZ(p)*pow(v_d+a_d, 2))*integral(h8, x, y);
      }

      template<>
        double h<9>(double x, double y, parameters const & p){
        return (CAA(p)*2*e_d+CAZ(p)*(-2*(v_d+a_d+e_d*sqrt(cw2/sw2)))+CZZ(p)*2*sqrt(cw2/sw2)*(v_d+a_d))*integral(h9, x, y);
      }

      template<>
        double h<10>(double x, double y, parameters const & p){
        return (CAA(p)+CAZ(p)*(-2*sqrt(cw2/sw2))+CZZ(p)*cw2/sw2)*integral(h10, x, y);
      }

      template<order O>
        double R_nonres(double x, double y, parameters const & p);

      //! NLO non-resonant contribution to \f$R_t\f$
      /**
       * cf. eq. (13) of \cite Beneke:2010mp
       * but normalised to the R ratio
       */
      template<>
      double R_nonres<NLO>(double x, double y, parameters const & p){
        return 24*pi*p.width/p.m*sum_h<1, 10>{}(x, y, p);
      }

      //! NNLO non-resonant contribution to \f$R_t\f$
      template<>
      double R_nonres<NNLO>(double x, double y, parameters const & p){
        assert((y <= 1) && (y > x));
        const double yw = (1-y)/(1-x);
        const double s = p.expand_s?(4*p.m*p.m):p.s;
        const double Cv = Wilson_coefficients::Cv_p(p);
        const double Ca = Wilson_coefficients::Ca_p(p);
        const double Cv_P_wave = Wilson_coefficients::Cv_p_P_wave(p);
        const double Ca_P_wave = Wilson_coefficients::Ca_p_P_wave(p);
        using Wilson_coefficients::Ca_abs;
        using Wilson_coefficients::Cv_abs;
        using Wilson_coefficients::C_abs_tag;
        return p.width*p.alpha_s*(
            + NNLO_nonres_interpolation(manual, x, yw)
            + NNLO_nonres_interpolation(automated, x, yw)
            + 6.*QCD::n_c*QCD::c_F*p.m*log(p.m/p.mu_w)*(
                Cv*Cv + Ca*Ca + Cv_P_wave*Cv_P_wave + Ca_P_wave*Ca_P_wave
                + 1/(4.*pi)*p.alpha*p.m/p.width*(
                    // beware!
                    // different prefactors due to different conventions
                    // w.r.t to the dimensionality of loop integrals
                    + Cv*(2.*Cv_abs<2, C_abs_tag::bare>(p) + 3.*Cv_abs<2, C_abs_tag::Zt>(p))
                    + Ca*(2.*Ca_abs<2, C_abs_tag::bare>(p) + 3.*Ca_abs<2, C_abs_tag::Zt>(p))
                )
            )/s
        );
      }

    }

    //! Computes the non-resonant contribution to \f$\sigma(e^+e^- \to t\bar{t})\f$
    /**
     * cf. eq. (13) of \cite Beneke:2010mp
     * but normalised to the R ratio
     */
    double R_nonres(parameters const & p){
      static constexpr double ep = 1e-10;
      const double x = pow(mW/p.m, 2);
      if(p.invariant_mass_cut < 0){
        throw std::invalid_argument("Negative invariant mass cut");
      }
      const double max_mass_cut = p.m*(1 - sqrt(x));
      if(p.invariant_mass_cut > max_mass_cut*(1+ep)){
        throw std::invalid_argument(
            "Invariant mass cut too large: can be at most "
            + std::to_string(max_mass_cut)
        );
      }
      double y = pow(1 - p.invariant_mass_cut/p.m, 2);
      // ensure we are in the grid range
      if(y <= x*(1+ep)) y = x*(1+ep);
      double result = 0;
      assert(p.order > 0);
      if(p.contributions.nonresonant[0]){
        result += p.contributions.nonresonant[0]*R_nonres<NLO>(x, y, p);
      }
      if(p.order >= 2 && p.contributions.nonresonant[1]){
        /*
         * we multiplied the NLO contribution by the all-order width
         * instead of just the LO width. To correct for this, we now
         * have to subtract the contribution from the NLO correction
         * to the width again.
         */
        const double c = p.contributions.nonresonant[1];
        result *= 1 - c*p.alpha_s/pi*x_1(sqrt(x));
        result += c*R_nonres<NNLO>(x, y, p);
      }
      return result;
    }
  }
}
