#include "QQbar_threshold/structure_function.hpp"

#include <cmath>

#include "QQbar_threshold/parameters.hpp"
#include "QQbar_threshold/constants.hpp"
#include "QQbar_threshold/zeta.hpp"
#include "QQbar_threshold/Gamma.hpp"
#include "QQbar_threshold/dilog.hpp"
#include "QQbar_threshold/hypergeometric.hpp"
#include "QQbar_threshold/integrate.hpp"

using zeta::pi;
using zeta::gamma_Euler;
using dilog::Li2;
namespace QQbar_threshold{
  using Gamma::Gamma;
  using hypergeometric::PFQ;

  namespace{
    // from eq. (67) of hep-ph/9602351
    double f_sing(double x, double beta){
      return beta*std::exp((3./8. - gamma_Euler/2)*beta)/
        (2.*Gamma(1. + beta/2.)*std::pow(1. - x, 1. - beta/2.));
    }
    double f_reg(double x, double beta){
      return -1./4.*beta*(
          + 1. + x
          + beta/8.*(
              5. + x + 4.*(1.+x)*log(1.-x) + (1.+3.*x*x)*log(x)/(1.-x)
              + beta/12.*(
                  (
                      + 39./4. + 15./4.*x + 6.*(5.+x)*log(1.-x)
                      + log(x)*(
                          + 3./2. + x*(12. + 9./2.*x)
                          + 12.*(1 + x*x)*log(1.-x)
                          - (1./2. + 7./2.*x*x)*log(x)
                      )/(1.-x)
                      + (1.+x)*(-3.*pi*pi + 12.*log(1.-x)*log(1.-x) + 6*Li2(x))
                  )
              )
          )
      );
    }

    double L_sing(double x, double beta){
      return exp((3./4.-gamma_Euler)*beta)*pow(1.-x, beta-1.)/Gamma(beta);
    }

    double L_int(double x, double beta){
      const auto integrand = [x,beta](double y){
        const double f_reg_xy = f_reg(x/y, beta)/y;
        return
        + 2*f_sing(y, beta)*(f_reg_xy - f_reg(x, beta))
        + f_reg(y, beta)*f_reg_xy;
      };
      return integrate(integrand, x , 1);
    }

    double L_reg(double x, double beta){
      const double f_reg_x = f_reg(x, beta);
      return
        + L_sing(x, beta)*(
            // rewritten to allow numerical Mellin-Barnes integration
            // originally: PFQ<2>({beta/2, beta/2}, {beta}, 1.-x) - 1.
            pow(x, -beta/2)*PFQ<2>({beta/2, beta/2}, {beta}, 1.-1./x) - 1.
        )
        + 2*f_reg_x*exp((3./8. - gamma_Euler/2)*beta)*pow(1.-x, beta/2.)
        /Gamma(1.+beta/2.)
        + L_int(x, beta)
        ;
    }

  }

  double modified_structure_function(double t, double beta){
    const double t2b = std::pow(t, 2/beta);
    const double log_t2b = std::log(t2b);
    const double log_x = std::log(1-t2b);
    return
      std::exp((3./8. - gamma_Euler/2)*beta)/Gamma(1. + beta/2.)
      + (
          -3.*t2b*(
              256 + (-128 + beta*(-16 + beta*(-5 + 4*pi*pi)))*t2b
              + beta*(
                  96 - 8*(-2*(8 + 3*beta) + (8 + beta)*t2b)*log_t2b
                  + beta*(18 - 8*pi*pi - 16*(-2 + t2b)*log_t2b*log_t2b)
              )
          )
          + beta*(
              log_x*(
                  2*beta*(8 + t2b*(-14 + 7*t2b))*log_x
                  - 6*(
                      4*(8 + 3*beta)
                      + t2b*(-2*(24 + 7*beta) + 3*(8 + beta)*t2b)
                      + 8*beta*(2 + (-2 + t2b)*t2b)*log_t2b
                  )
              )
              + 24*beta*(-2 + t2b)*t2b*Li2(1 - t2b)
          )
      )/(768*t);
  }

  double modified_luminosity_function(double t, double beta){
    const double x = 1 - pow(t, 1./beta);
    const double f_reg_x = f_reg(x, beta);
    const double small_t_exp = exp((3./4.-gamma_Euler)*beta)/(beta*Gamma(beta));
    if(t < 0.01) return small_t_exp;
    return
      // rewritten to allow numerical Mellin-Barnes integration
      // originally: small_t_exp*PFQ<2>({beta/2, beta/2}, {beta}, 1.-x)
      small_t_exp*pow(x, -beta/2)*PFQ<2>({beta/2, beta/2}, {beta}, 1.-1./x)
      + 2*f_reg_x*exp((3./8. - gamma_Euler/2)*beta)*pow(1.-x, 1-beta/2.)
      /(beta*Gamma(1.+beta/2.))
      + pow(1.-x, 1-beta)/beta*L_int(x, beta)
      ;
  }

  double luminosity_function(double x, double beta){
    return L_sing(x, beta) + L_reg(x, beta);
  }

  double structure_function(double x, double beta){
    return f_sing(x, beta) + f_reg(x, beta);
  }

  double ISR_log(double sqrt_s, double alpha){
    return -2./pi*alpha*(2*std::log(me/sqrt_s) + 1.);
  }

  double top_ISR_log(double sqrt_s){
    return ISR_log(sqrt_s, alpha_mZ);
  }

  double bottom_ISR_log(double sqrt_s){
    return ISR_log(sqrt_s, alpha_Y);
  }

}
