#ifdef NDEBUG
#undef NDEBUG
#endif

#include <complex>
#include <cassert>
#include <array>
#include <iostream>

#include "energy_levels.hpp"
#include "mass_schemes.hpp"
#include "scheme_conversion.hpp"
#include "constants.hpp"

#define TSTB(name,i) \
  std::cout << #name << ": "; \
  diff = name - results_bottom[i]; \
  std::cout << diff; \
  assert(std::abs(diff)/std::abs(results_bottom[i]) < ep);	\
  std::cout << " (ok)\n";

#define TSTT(name,i) \
  std::cout << #name << ": "; \
  diff = name - results_top[i]; \
  std::cout << diff; \
  assert(std::abs(diff)/std::abs(results_top[i]) < ep);	\
  std::cout << " (ok)\n";

int main(){
  using namespace QQbar_threshold::pert_order;
  using QQbar_threshold::Pole;
  using QQbar_threshold::PS;
  using QQbar_threshold::PSshift;
  using QQbar_threshold::mu_f_top;
  using QQbar_threshold::mu_f_bottom;
  using QQbar_threshold::top_options;
  using QQbar_threshold::bottom_options;
  using QQbar_threshold::SM;
  QQbar_threshold::options opt = top_options();
  opt.alpha = 1/137.03599907430635;
  opt.m_Higgs = 125.9;
  opt.ml = 0;
  opt.beyond_QCD = SM::none;
  using QQbar_threshold::bbbar_energy_level;
  using QQbar_threshold::ttbar_energy_level;
  double diff;
  double constexpr ep = 1e-3;
  std::cout << "target precision: " << ep << "\n";

  std::cout << "--------------------------------------------------\n";
  std::cout << "top, pole scheme:\n";

  opt.mass_scheme = {Pole, 0};

  auto results_top = std::array<double, 4>{{
      -1.1349849845898385,
      -1.922869905087344,
      -2.4434663129349685,
      -2.7594768627724764
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTT(ttbar_energy_level(1, 80, 175, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "top, PS scheme:\n";

  opt.mass_scheme = {PS, mu_f_top};

  results_top = std::array<double, 4>{{
      0.9157812013604718,
      0.9434154257194796,
      0.8452081704684791,
      0.7969198922604583
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTT(ttbar_energy_level(1, 80, 175, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "top, PSshift scheme:\n";

  opt.mass_scheme = {PSshift, mu_f_top};

  results_top = std::array<double, 4>{{
      0.9091309475663056,
      0.9304195841171294,
      0.8349937194221653,
      0.7902051037056448
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTT(ttbar_energy_level(1, 80, 175, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "top, QED effects:\n";
  opt.beyond_QCD = SM::QED;

  results_top = std::array<double, 4>{{
      0.9091309475663056,
      0.8843367675587837,
      0.7792405318739175,
      0.7294959444611009
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTT(ttbar_energy_level(1, 80, 175, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "top, QED + Higgs effects:\n";
  opt.beyond_QCD |= SM::Higgs ;

  results_top = std::array<double, 4>{{
      0.9091309475663056,
      0.8843367675587837,
      0.7792405318739175,
      0.7006686292302344
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTT(ttbar_energy_level(1, 80, 175, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "top, width effects:\n";

  constexpr double ref_deltaE_width = 0.00255403;
  diff =
    ttbar_energy_level(1, 80, {171.5, 1.33}, N2LO, opt)
    - ttbar_energy_level(1, 80, 171.5, N2LO, opt)
    - ref_deltaE_width;
  std::cout << diff << '\n';
  assert(std::abs(diff) < 1e-8);

  std::cout << "--------------------------------------------------\n";
  std::cout << "bottom, Pole scheme:\n";
  opt = bottom_options();
  opt.beyond_QCD = SM::none;
  opt.alpha_s_mZ = 0.11842395335727579;

  opt.mass_scheme = {Pole, 0};

  auto results_bottom = std::array<double, 4>{{
      -0.025353659373968367,
      -0.07998161061009382,
      -0.17430060256872895,
      -0.3293001620203085
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTB(bbbar_energy_level(2, 5, 5, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "bottom, PS scheme:\n";

  opt.mass_scheme = {PS, mu_f_bottom};

  results_bottom = std::array<double,4>{{
      0.3373111641265726,
      0.5158973240857477,
      0.6388514363897094,
      0.7363973715089671
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTB(bbbar_energy_level(2, 5, 5, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "bottom, PSshift scheme:\n";

  opt.mass_scheme = {PSshift, mu_f_bottom};

  results_bottom = std::array<double,4>{{
      0.33639167608637727,
      0.5120125022207253,
      0.6304697913782349,
      0.7215210926478872
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTB(bbbar_energy_level(2, 5, 5, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "bottom, PSshift scheme (mc = 0.9):\n";

  opt.mass_scheme = {PSshift, mu_f_bottom};
  opt.ml = 0.9;

  results_bottom = std::array<double,4>{{
      0.33639167608637727,
      0.5183861965190428,
      0.6498165693971962,
      0.7405376837722965
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTB(bbbar_energy_level(2, 5, 5, o, opt), o);
  }


  std::cout << "--------------------------------------------------\n";
  std::cout << "bottom, PSshift scheme (mc = 0.9) with massive double insertions:\n";

  QQbar_threshold::options opt2 = opt;
  opt2.double_light_insertion = true;

  results_bottom = std::array<double,4>{{
      0,
      0,
      -0.0015216058107720698,
      -0.0014901852000416271
    }};

  for(order o = LO; o <= N3LO; ++o){
    double diff;
    diff = std::abs(
	bbbar_energy_level(2, 5, 5, o, opt2)
	- bbbar_energy_level(2, 5, 5, o, opt)
	- results_bottom[o]
    );
    std::cout << o << ": " << diff;
    assert(diff < ep);
    std::cout << " (ok)\n";
  }

}
